import glob
import json
import numpy as np
import os
import random
import shutil
import tempfile
import unittest

import ray
from ray.tune.registry import (
    register_env,
    register_input,
    registry_get_input,
    registry_contains_input,
)
from ray.rllib.algorithms.algorithm_config import AlgorithmConfig
from ray.rllib.algorithms.pg import PGConfig
from ray.rllib.examples.env.multi_agent import MultiAgentCartPole
from ray.rllib.offline import (
    IOContext,
    JsonWriter,
    JsonReader,
    InputReader,
    ShuffledInput,
    DatasetWriter,
)
from ray.rllib.offline.json_reader import from_json_data
from ray.rllib.offline.json_writer import _to_json_dict, _to_json
from ray.rllib.policy.sample_batch import SampleBatch, convert_ma_batch_to_sample_batch
from ray.rllib.utils.test_utils import framework_iterator

SAMPLES = SampleBatch(
    {
        "actions": np.array([1, 2, 3, 4]),
        "obs": np.array([4, 5, 6, 7]),
        "eps_id": [1, 1, 2, 3],
    }
)


def make_sample_batch(i):
    return SampleBatch({"actions": np.array([i, i, i]), "obs": np.array([i, i, i])})


class AgentIOTest(unittest.TestCase):
    def setUp(self):
        ray.init(num_cpus=1, ignore_reinit_error=True)
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        ray.shutdown()

    def write_outputs(self, output, fw, output_config=None):
        config = (
            PGConfig()
            .environment("CartPole-v1")
            .framework(fw)
            .training(train_batch_size=250)
            .offline_data(
                output=output + (fw if output != "logdir" else ""),
                output_config=output_config or {},
            )
        )
        algo = config.build()
        algo.train()
        return algo

    def test_agent_output_ok(self):
        for fw in framework_iterator(frameworks=("torch", "tf")):
            self.write_outputs(self.test_dir, fw)
            self.assertEqual(len(os.listdir(self.test_dir + fw)), 1)
            reader = JsonReader(self.test_dir + fw + "/*.json")
            reader.next()

    def test_agent_output_logdir(self):
        """Test special value 'logdir' as Agent's output."""
        for fw in framework_iterator():
            agent = self.write_outputs("logdir", fw)
            self.assertEqual(len(glob.glob(agent.logdir + "/output-*.json")), 1)

    def test_agent_output_infos(self):
        """Verify that the infos dictionary is written to the output files.

        Note, with torch this is always the case.
        """
        output_config = {"store_infos": True}
        for fw in framework_iterator(frameworks=("torch", "tf")):
            self.write_outputs(self.test_dir, fw, output_config=output_config)
            self.assertEqual(len(os.listdir(self.test_dir + fw)), 1)
            reader = JsonReader(self.test_dir + fw + "/*.json")
            data = reader.next()
            data = convert_ma_batch_to_sample_batch(data)
            assert "infos" in data

    def test_agent_input_dir(self):
        config = (
            PGConfig()
            .environment("CartPole-v1")
            .evaluation(off_policy_estimation_methods={})
        )

        for fw in framework_iterator(config, frameworks=("torch", "tf")):
            self.write_outputs(self.test_dir, fw)
            config.offline_data(
                input_=self.test_dir + fw,
            )
            print("WROTE TO: ", self.test_dir)
            algo = config.build()
            result = algo.train()
            self.assertEqual(result["timesteps_total"], 250)  # read from input
            self.assertTrue(np.isnan(result["episode_reward_mean"]))

    def test_split_by_episode(self):
        splits = SAMPLES.split_by_episode()
        self.assertEqual(len(splits), 3)
        self.assertEqual(splits[0].count, 2)
        self.assertEqual(splits[1].count, 1)
        self.assertEqual(splits[2].count, 1)

    def test_agent_input_postprocessing_enabled(self):
        config = (
            PGConfig()
            .environment("CartPole-v1")
            .offline_data(
                postprocess_inputs=True,  # adds back 'advantages'
            )
            .evaluation(off_policy_estimation_methods={})
        )

        for fw in framework_iterator(config, frameworks=("tf", "torch")):
            self.write_outputs(self.test_dir, fw)
            config.offline_data(input_=self.test_dir + fw)

            # Rewrite the files to drop advantages and value_targets for
            # testing
            for path in glob.glob(self.test_dir + fw + "/*.json"):
                out = []
                with open(path) as f:
                    for line in f.readlines():
                        data_string = json.loads(line)
                        data = from_json_data(data_string, None)
                        data = convert_ma_batch_to_sample_batch(data)
                        # Data won't contain rewards as these are not included
                        # in the write_outputs run (not needed in the
                        # SampleBatch). Flip out "rewards" for "advantages"
                        # just for testing.
                        data["rewards"] = data["advantages"]
                        del data["advantages"]
                        if "value_targets" in data:
                            del data["value_targets"]
                        out.append(_to_json_dict(data, []))
                with open(path, "w") as f:
                    for data in out:
                        f.write(json.dumps(data))

            algo = config.build()
            result = algo.train()
            self.assertEqual(result["timesteps_total"], 250)  # read from input
            self.assertTrue(np.isnan(result["episode_reward_mean"]))
            algo.stop()

    def test_agent_input_eval_sampler(self):
        config = (
            PGConfig()
            .environment("CartPole-v1")
            .offline_data(
                postprocess_inputs=True,  # adds back 'advantages'
            )
            .evaluation(
                evaluation_interval=1,
                evaluation_config=PGConfig.overrides(input_="sampler"),
            )
        )

        for fw in framework_iterator(config, frameworks=["tf", "torch"]):
            self.write_outputs(self.test_dir, fw)
            config.offline_data(input_=self.test_dir + fw)
            algo = config.build()
            result = algo.train()
            assert np.isnan(
                result["episode_reward_mean"]
            ), "episode reward should not be computed for offline data"
            assert not np.isnan(
                result["evaluation"]["episode_reward_mean"]
            ), "Did not see simulation results during evaluation"
            algo.stop()

    def test_agent_input_list(self):
        config = (
            PGConfig()
            .environment("CartPole-v1")
            .training(train_batch_size=99)
            .evaluation(off_policy_estimation_methods={})
        )

        for fw in framework_iterator(config, frameworks=("torch", "tf")):
            self.write_outputs(self.test_dir, fw)
            config.offline_data(input_=glob.glob(self.test_dir + fw + "/*.json"))
            algo = config.build()
            result = algo.train()
            self.assertEqual(result["timesteps_total"], 250)  # read from input
            self.assertTrue(np.isnan(result["episode_reward_mean"]))
            algo.stop()

    def test_agent_input_dict(self):
        config = PGConfig().environment("CartPole-v1").training(train_batch_size=2000)
        for fw in framework_iterator(config):
            self.write_outputs(self.test_dir, fw)
            config.offline_data(
                input_={
                    self.test_dir + fw: 0.1,
                    "sampler": 0.9,
                }
            )
            algo = config.build()
            result = algo.train()
            self.assertTrue(not np.isnan(result["episode_reward_mean"]))
            algo.stop()

    def test_multi_agent(self):
        register_env(
            "multi_agent_cartpole", lambda _: MultiAgentCartPole({"num_agents": 10})
        )

        config = (
            PGConfig()
            .environment("multi_agent_cartpole")
            .rollouts(num_rollout_workers=0)
            .multi_agent(
                policies={"policy_1", "policy_2"},
                policy_mapping_fn=lambda agent_id, episode, worker, **kwargs: (
                    random.choice(["policy_1", "policy_2"])
                ),
            )
        )

        for fw in framework_iterator(config):
            config.offline_data(output=self.test_dir + fw)
            pg = config.build()
            pg.train()
            self.assertEqual(len(os.listdir(self.test_dir + fw)), 1)
            pg.stop()

            config2 = config.copy()
            config2.output = None
            config2.evaluation(
                evaluation_interval=1,
                evaluation_config=PGConfig.overrides(input_="sampler"),
            )
            config2.training(train_batch_size=2000)
            config2.offline_data(input_=self.test_dir + fw)

            pg2 = config2.build()
            result = pg2.train()
            assert np.isnan(
                result["episode_reward_mean"]
            ), "episode reward should not be computed for offline data"
            assert not np.isnan(
                result["evaluation"]["episode_reward_mean"]
            ), "Did not see simulation results during evaluation"
            pg2.stop()

    def test_custom_input_procedure(self):
        class CustomJsonReader(JsonReader):
            def __init__(self, ioctx: IOContext):
                super().__init__(ioctx.input_config["input_files"], ioctx)

        def input_creator(ioctx: IOContext) -> InputReader:
            return ShuffledInput(CustomJsonReader(ioctx))

        register_input("custom_input", input_creator)
        test_input_procedure = [
            "custom_input",
            input_creator,
            "ray.rllib.examples.custom_input_api.CustomJsonReader",
        ]

        for input_procedure in test_input_procedure:

            config = (
                PGConfig()
                .environment("CartPole-v1")
                .offline_data(input_=input_procedure)
                .evaluation(off_policy_estimation_methods={})
            )

            for fw in framework_iterator(config, frameworks=("torch", "tf")):
                self.write_outputs(self.test_dir, fw)
                config.offline_data(input_config={"input_files": self.test_dir + fw})
                algo = config.build()
                result = algo.train()
                self.assertEqual(result["timesteps_total"], 250)
                self.assertTrue(np.isnan(result["episode_reward_mean"]))
                algo.stop()

    def test_multiple_output_workers(self):
        ray.shutdown()
        ray.init(num_cpus=4, ignore_reinit_error=True)

        config = (
            PGConfig()
            .environment("CartPole-v1")
            .rollouts(num_rollout_workers=2)
            .training(train_batch_size=500)
            .evaluation(off_policy_estimation_methods={})
        )

        for fw in framework_iterator(config, frameworks=["tf", "torch"]):
            config.offline_data(output=self.test_dir + fw)
            algo = config.build()
            algo.train()
            self.assertEqual(len(os.listdir(self.test_dir + fw)), 2)
            reader = JsonReader(self.test_dir + fw + "/*.json")
            reader.next()
            algo.stop()


class JsonIOTest(unittest.TestCase):
    def setUp(self):
        ray.init(num_cpus=1, ignore_reinit_error=True)
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        ray.shutdown()

    def test_write_dataset(self):
        ioctx = IOContext(
            self.test_dir,
            AlgorithmConfig().offline_data(
                output="dataset",
                output_config={
                    "format": "json",
                    "path": self.test_dir,
                    "max_num_samples_per_file": 2,
                },
            ),
            0,
            None,
        )
        writer = DatasetWriter(ioctx, compress_columns=["obs"])
        self.assertEqual(len(os.listdir(self.test_dir)), 0)
        writer.write(SAMPLES)
        writer.write(SAMPLES)
        self.assertEqual(len(os.listdir(self.test_dir)), 1)

    def test_write_simple(self):
        ioctx = IOContext(self.test_dir, {}, 0, None)
        writer = JsonWriter(
            self.test_dir, ioctx, max_file_size=1000, compress_columns=["obs"]
        )
        self.assertEqual(len(os.listdir(self.test_dir)), 0)
        writer.write(SAMPLES)
        writer.write(SAMPLES)
        self.assertEqual(len(os.listdir(self.test_dir)), 1)

    def test_write_file_uri(self):
        ioctx = IOContext(self.test_dir, None, 0, None)
        writer = JsonWriter(
            "file://" + self.test_dir,
            ioctx,
            max_file_size=1000,
            compress_columns=["obs"],
        )
        self.assertEqual(len(os.listdir(self.test_dir)), 0)
        writer.write(SAMPLES)
        writer.write(SAMPLES)
        self.assertEqual(len(os.listdir(self.test_dir)), 1)

    def test_write_paginate(self):
        ioctx = IOContext(self.test_dir, AlgorithmConfig(), 0, None)
        writer = JsonWriter(
            self.test_dir, ioctx, max_file_size=5000, compress_columns=["obs"]
        )
        self.assertEqual(len(os.listdir(self.test_dir)), 0)
        for _ in range(100):
            writer.write(SAMPLES)
        num_files = len(os.listdir(self.test_dir))

        # Pagination can't really be predicted:
        # On travis, it seems to create only 2 files, but sometimes also
        # 6, or 7. 12 or 13 usually on a Mac locally.
        # Reasons: Different compressions, file-size interpretations,
        # json writers?
        assert num_files >= 2, "Expected >= 2 files, but found {} ({})".format(
            num_files, os.listdir(self.test_dir)
        )

    def test_read_write(self):
        ioctx = IOContext(self.test_dir, None, 0, None)
        writer = JsonWriter(
            self.test_dir, ioctx, max_file_size=5000, compress_columns=["obs"]
        )
        for i in range(100):
            writer.write(make_sample_batch(i))
        reader = JsonReader(self.test_dir + "/*.json")
        seen_a = set()
        seen_o = set()
        for i in range(1000):
            batch = reader.next()
            seen_a.add(batch["actions"][0])
            seen_o.add(batch["obs"][0])
        self.assertGreater(len(seen_a), 90)
        self.assertLess(len(seen_a), 101)
        self.assertGreater(len(seen_o), 90)
        self.assertLess(len(seen_o), 101)

    def test_skips_over_empty_lines_and_files(self):
        open(self.test_dir + "/empty", "w").close()
        with open(self.test_dir + "/f1", "w") as f:
            f.write("\n")
            f.write("\n")
            f.write(_to_json(make_sample_batch(0), []))
        with open(self.test_dir + "/f2", "w") as f:
            f.write(_to_json(make_sample_batch(1), []))
            f.write("\n")
        reader = JsonReader(
            [
                self.test_dir + "/empty",
                self.test_dir + "/f1",
                "file://" + self.test_dir + "/f2",
            ]
        )
        seen_a = set()
        for i in range(100):
            batch = reader.next()
            seen_a.add(batch["actions"][0])
        self.assertEqual(len(seen_a), 2)

    def test_skips_over_corrupted_lines(self):
        with open(self.test_dir + "/f1", "w") as f:
            f.write(_to_json(make_sample_batch(0), []))
            f.write("\n")
            f.write(_to_json(make_sample_batch(1), []))
            f.write("\n")
            f.write(_to_json(make_sample_batch(2), []))
            f.write("\n")
            f.write(_to_json(make_sample_batch(3), []))
            f.write("\n")
            f.write("{..corrupted_json_record")
        reader = JsonReader(
            [
                self.test_dir + "/f1",
            ]
        )
        seen_a = set()
        for i in range(10):
            batch = reader.next()
            seen_a.add(batch["actions"][0])
        self.assertEqual(len(seen_a), 4)

    def test_abort_on_all_empty_inputs(self):
        open(self.test_dir + "/empty", "w").close()
        reader = JsonReader(
            [
                self.test_dir + "/empty",
            ]
        )
        self.assertRaises(ValueError, lambda: reader.next())
        with open(self.test_dir + "/empty1", "w") as f:
            for _ in range(100):
                f.write("\n")
        with open(self.test_dir + "/empty2", "w") as f:
            for _ in range(100):
                f.write("\n")
        reader = JsonReader(
            [
                self.test_dir + "/empty1",
                self.test_dir + "/empty2",
            ]
        )
        self.assertRaises(ValueError, lambda: reader.next())

    def test_custom_input_registry(self):
        config = AlgorithmConfig().offline_data(input_config={})
        ioctx = IOContext(self.test_dir, config, 0, None)

        class CustomInputReader(InputReader):
            def __init__(self, ioctx: IOContext):
                self.ioctx = ioctx

            def next(self):
                return 0

        def input_creator(ioctx: IOContext):
            return ShuffledInput(CustomInputReader(ioctx))

        register_input("custom_input", input_creator)
        self.assertTrue(registry_contains_input("custom_input"))
        creator = registry_get_input("custom_input")
        self.assertIsNotNone(creator)
        reader = creator(ioctx)
        self.assertIsInstance(reader, ShuffledInput)
        self.assertEqual(reader.next(), 0)
        self.assertEqual(ioctx.log_dir, self.test_dir)
        self.assertEqual(ioctx.config, config)
        self.assertEqual(ioctx.worker_index, 0)
        self.assertIsNone(ioctx.worker)
        self.assertEqual(ioctx.input_config, {})


if __name__ == "__main__":
    import pytest
    import sys

    sys.exit(pytest.main(["-v", __file__]))



# Adapted from https://github.com/pyg-team/pytorch_geometric/blob/2.1.0
# /examples/multi_gpu/distributed_sampling.py

import os
import argparse
from filelock import FileLock
from ray.air import session

import torch
import torch.nn.functional as F

from torch_geometric.datasets import Reddit, FakeDataset
from torch_geometric.loader import NeighborSampler
from torch_geometric.nn import SAGEConv

from ray import train
from ray.train.torch import TorchTrainer
from ray.air.config import ScalingConfig
from torch_geometric.transforms import RandomNodeSplit


class SAGE(torch.nn.Module):
    def __init__(self, in_channels, hidden_channels, out_channels, num_layers=2):
        super().__init__()
        self.num_layers = num_layers

        self.convs = torch.nn.ModuleList()
        self.convs.append(SAGEConv(in_channels, hidden_channels))
        for _ in range(self.num_layers - 2):
            self.convs.append(SAGEConv(hidden_channels, hidden_channels))
        self.convs.append(SAGEConv(hidden_channels, out_channels))

    def forward(self, x, adjs):
        for i, (edge_index, _, size) in enumerate(adjs):
            x_target = x[: size[1]]  # Target nodes are always placed first.
            x = self.convs[i]((x, x_target), edge_index)
            if i != self.num_layers - 1:
                x = F.relu(x)
                x = F.dropout(x, p=0.5, training=self.training)
        return x.log_softmax(dim=-1)

    @torch.no_grad()
    def test(self, x_all, subgraph_loader):
        for i in range(self.num_layers):
            xs = []
            for batch_size, n_id, adj in subgraph_loader:

                edge_index, _, size = adj
                x = x_all[n_id.to(x_all.device)].to(train.torch.get_device())
                x_target = x[: size[1]]
                x = self.convs[i]((x, x_target), edge_index)
                if i != self.num_layers - 1:
                    x = F.relu(x)
                xs.append(x.cpu())

            x_all = torch.cat(xs, dim=0)

        return x_all


def train_loop_per_worker(train_loop_config):
    dataset = train_loop_config["dataset_fn"]()
    batch_size = train_loop_config["batch_size"]
    num_epochs = train_loop_config["num_epochs"]

    data = dataset[0]
    train_idx = data.train_mask.nonzero(as_tuple=False).view(-1)
    train_idx = train_idx.split(train_idx.size(0) // session.get_world_size())[
        session.get_world_rank()
    ]

    train_loader = NeighborSampler(
        data.edge_index,
        node_idx=train_idx,
        sizes=[25, 10],
        batch_size=batch_size,
        shuffle=True,
    )

    # Disable distributed sampler since the train_loader has already been split above.
    train_loader = train.torch.prepare_data_loader(train_loader, add_dist_sampler=False)

    # Do validation on rank 0 worker only.
    if session.get_world_rank() == 0:
        subgraph_loader = NeighborSampler(
            data.edge_index, node_idx=None, sizes=[-1], batch_size=2048, shuffle=False
        )
        subgraph_loader = train.torch.prepare_data_loader(
            subgraph_loader, add_dist_sampler=False
        )

    model = SAGE(dataset.num_features, 256, dataset.num_classes)
    model = train.torch.prepare_model(model)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)

    x, y = data.x.to(train.torch.get_device()), data.y.to(train.torch.get_device())

    for epoch in range(num_epochs):
        model.train()

        # ``batch_size`` is the number of samples in the current batch.
        # ``n_id`` are the ids of all the nodes used in the computation. This is
        # needed to pull in the necessary features just for the current batch that is
        # being trained on.
        # ``adjs`` is a list of 3 element tuple consisting of ``(edge_index, e_id,
        # size)`` for each sample in the batch, where ``edge_index``represent the
        # edges of the sampled subgraph, ``e_id`` are the ids of the edges in the
        # sample, and ``size`` holds the shape of the subgraph.
        # See ``torch_geometric.loader.neighbor_sampler.NeighborSampler`` for more info.
        for batch_size, n_id, adjs in train_loader:
            optimizer.zero_grad()
            out = model(x[n_id], adjs)
            loss = F.nll_loss(out, y[n_id[:batch_size]])
            loss.backward()
            optimizer.step()

        if session.get_world_rank() == 0:
            print(f"Epoch: {epoch:03d}, Loss: {loss:.4f}")

        train_accuracy = validation_accuracy = test_accuracy = None

        # Do validation on rank 0 worker only.
        if session.get_world_rank() == 0:
            model.eval()
            with torch.no_grad():
                out = model.module.test(x, subgraph_loader)
            res = out.argmax(dim=-1) == data.y
            train_accuracy = int(res[data.train_mask].sum()) / int(
                data.train_mask.sum()
            )
            validation_accuracy = int(res[data.val_mask].sum()) / int(
                data.val_mask.sum()
            )
            test_accuracy = int(res[data.test_mask].sum()) / int(data.test_mask.sum())

        session.report(
            dict(
                train_accuracy=train_accuracy,
                validation_accuracy=validation_accuracy,
                test_accuracy=test_accuracy,
            )
        )


def gen_fake_dataset():
    """Returns a function to be called on each worker that returns a Fake Dataset."""

    # For fake dataset, since the dataset is randomized, we create it once on the
    # driver, and then send the same dataset to all the training workers.
    # Use 10% of nodes for validation and 10% for testing.
    fake_dataset = FakeDataset(transform=RandomNodeSplit(num_val=0.1, num_test=0.1))

    def gen_dataset():
        return fake_dataset

    return gen_dataset


def gen_reddit_dataset():
    """Returns a function to be called on each worker that returns Reddit Dataset."""

    # For Reddit dataset, we have to download the data on each node, so we create the
    # dataset on each training worker.
    with FileLock(os.path.expanduser("~/.reddit_dataset_lock")):
        dataset = Reddit("./data/Reddit")
    return dataset


def train_gnn(
    num_workers=2, use_gpu=False, epochs=3, global_batch_size=32, dataset="reddit"
):

    per_worker_batch_size = global_batch_size // num_workers

    trainer = TorchTrainer(
        train_loop_per_worker=train_loop_per_worker,
        train_loop_config={
            "num_epochs": epochs,
            "batch_size": per_worker_batch_size,
            "dataset_fn": gen_reddit_dataset
            if dataset == "reddit"
            else gen_fake_dataset(),
        },
        scaling_config=ScalingConfig(num_workers=num_workers, use_gpu=use_gpu),
    )
    result = trainer.fit()
    print(result.metrics)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--address", required=False, type=str, help="the address to use for Ray"
    )
    parser.add_argument(
        "--num-workers",
        "-n",
        type=int,
        default=2,
        help="Sets number of workers for training.",
    )
    parser.add_argument(
        "--use-gpu", action="store_true", help="Whether to use GPU for training."
    )
    parser.add_argument(
        "--epochs", type=int, default=3, help="Number of epochs to train for."
    )
    parser.add_argument(
        "--global-batch-size",
        "-b",
        type=int,
        default=32,
        help="Global batch size to use for training.",
    )
    parser.add_argument(
        "--dataset",
        "-d",
        type=str,
        choices=["reddit", "fake"],
        default="reddit",
        help="The dataset to use. Either 'reddit' or 'fake' Defaults to 'reddit'.",
    )

    args, _ = parser.parse_known_args()

    train_gnn(
        num_workers=args.num_workers,
        use_gpu=args.use_gpu,
        epochs=args.epochs,
        global_batch_size=args.global_batch_size,
        dataset=args.dataset,
    )



import json
import os
from pathlib import Path

import ray
from ray import air, tune
from ray.tune.registry import get_trainable_cls

from ray.rllib.examples.env.stateless_cartpole import StatelessCartPole


param_space = {
    "num_gpus": int(os.environ.get("RLLIB_NUM_GPUS", "0")),
    "framework": "torch",
    "num_workers": 4,
    "num_envs_per_worker": 1,
    "num_cpus_per_worker": 1,
    "log_level": "INFO",
    "env": StatelessCartPole,
    "gamma": 0.95,
    "batch_mode": "complete_episodes",
    "replay_buffer_config": {
        "type": "MultiAgentReplayBuffer",
        "storage_unit": "sequences",
        "capacity": 100000,
        "replay_burn_in": 4,
    },
    "num_steps_sampled_before_learning_starts": 1000,
    "train_batch_size": 480,
    "target_network_update_freq": 480,
    "tau": 0.3,
    "zero_init_states": False,
    "optimization": {
        "actor_learning_rate": 0.005,
        "critic_learning_rate": 0.005,
        "entropy_learning_rate": 0.0001,
    },
    "model": {
        "max_seq_len": 20,
    },
    "policy_model_config": {
        "use_lstm": True,
        "lstm_cell_size": 64,
        "fcnet_hiddens": [64, 64],
        "lstm_use_prev_action": True,
        "lstm_use_prev_reward": True,
    },
    "q_model_config": {
        "use_lstm": True,
        "lstm_cell_size": 64,
        "fcnet_hiddens": [64, 64],
        "lstm_use_prev_action": True,
        "lstm_use_prev_reward": True,
    },
}

if __name__ == "__main__":
    # INIT
    ray.init(num_cpus=5)

    # TRAIN
    results = tune.Tuner(
        "RNNSAC",
        run_config=air.RunConfig(
            name="RNNSAC_example",
            local_dir=str(Path(__file__).parent / "example_out"),
            verbose=2,
            checkpoint_config=air.CheckpointConfig(
                checkpoint_at_end=True,
                num_to_keep=1,
                checkpoint_score_attribute="episode_reward_mean",
            ),
            stop={
                "episode_reward_mean": 65.0,
                "timesteps_total": 50000,
            },
        ),
        tune_config=tune.TuneConfig(
            metric="episode_reward_mean",
            mode="max",
        ),
        param_space=param_space,
    ).fit()

    # TEST

    checkpoint_config_path = os.path.join(
        results.get_best_result().log_dir, "params.json"
    )
    with open(checkpoint_config_path, "rb") as f:
        checkpoint_config = json.load(f)

    checkpoint_config["explore"] = False

    best_checkpoint = results.get_best_result().best_checkpoints[0][0]
    print("Loading checkpoint: {}".format(best_checkpoint))

    algo = get_trainable_cls("RNNSAC")(env=StatelessCartPole, config=checkpoint_config)
    algo.restore(best_checkpoint)

    env = algo.env_creator({})
    state = algo.get_policy().get_initial_state()
    prev_action = 0
    prev_reward = 0
    obs, info = env.reset()

    eps = 0
    ep_reward = 0
    while eps < 10:
        action, state, info_algo = algo.compute_single_action(
            obs,
            state=state,
            prev_action=prev_action,
            prev_reward=prev_reward,
            full_fetch=True,
        )
        obs, reward, terminated, truncated, info = env.step(action)
        prev_action = action
        prev_reward = reward
        ep_reward += reward
        try:
            env.render()
        except Exception:
            pass
        if terminated or truncated:
            eps += 1
            print("Episode {}: {}".format(eps, ep_reward))
            ep_reward = 0
            state = algo.get_policy().get_initial_state()
            prev_action = 0
            prev_reward = 0
            obs, info = env.reset()
    ray.shutdown()



import copy
from dataclasses import dataclass
import pprint
from typing import Iterator, Mapping, Any, Union, Dict, Optional, Type

from ray.util.annotations import PublicAPI
from ray.rllib.utils.annotations import override, ExperimentalAPI
from ray.rllib.utils.nested_dict import NestedDict

from ray.rllib.models.specs.specs_dict import SpecDict
from ray.rllib.policy.sample_batch import MultiAgentBatch
from ray.rllib.core.rl_module.rl_module import RLModule, SingleAgentRLModuleSpec

# TODO (Kourosh): change this to module_id later to enforce consistency
from ray.rllib.utils.policy import validate_policy_id

ModuleID = str


@ExperimentalAPI
@dataclass
class MultiAgentRLModuleSpec:
    """A utility spec class to make it constructing RLModules (in multi-agent case) easier.

    Args:
        module_class: ...
        module_specs: ...
    """

    module_class: Optional[Type["MultiAgentRLModule"]] = None
    module_specs: Optional[Dict[ModuleID, SingleAgentRLModuleSpec]] = None

    def build(self) -> "MultiAgentRLModule":
        return self.module_class.from_multi_agent_config({"modules": self.module_specs})


def _get_module_configs(config: Dict[str, Any]):
    """Constructs a mapping from module_id to module config.

    It takes care of the inheritance of common configs to individual module configs.
    See `from_multi_agent_config` for more details.
    """
    config = copy.deepcopy(config)
    module_specs = config.pop("modules", {})
    for common_spec in config:
        for module_spec in module_specs.values():
            if getattr(module_spec, common_spec) is None:
                setattr(module_spec, common_spec, config[common_spec])

    return module_specs


@PublicAPI(stability="alpha")
class MultiAgentRLModule(RLModule):
    """Base class for multi-agent RLModules.

    This class holds a mapping from module_ids to the underlying RLModules. It provides
    a convenient way of accessing each individual module, as well as accessing all of
    them with only one api call. Whether or not a given module is trainable is
    determined by the caller of this class (not the instance of this class itself).

    The extension of this class can include any arbitrary neural networks as part of
    the multi-agent module. For example, a multi-agent module can include a shared
    encoder network that is used by all the individual RLModules. It is up to the user
    to decide how to implement this class.

    The default implementation assumes the data communicated as input and output of
    the APIs in this class are `MultiAgentBatch` types. The `MultiAgentRLModule` simply
    loops through each `module_id`, and runs the forward pass of the corresponding
    `RLModule` object with the associated `SampleBatch` within the `MultiAgentBatch`.
    It also assumes that the underlying RLModules do not share any parameters or
    communication with one another. The behavior of modules with such advanced
    communication would be undefined by default. To share parameters or communication
    between the underlying RLModules, you should implement your own
    `MultiAgentRLModule`.
    """

    def __init__(self, rl_modules: Mapping[ModuleID, RLModule] = None) -> None:
        super().__init__()
        self._rl_modules: Mapping[ModuleID, RLModule] = rl_modules or {}

    @classmethod
    def from_multi_agent_config(cls, config: Mapping[str, Any]) -> "MultiAgentRLModule":
        """Creates a MultiAgentRLModule from a multi-agent config.

        The input config should contain "modules" key that is a mapping from module_id
        to the module spec for each RLModule which is a SingleAgentRLModuleSpec object.
        If there are multiple modules that do share the same
        `observation_space`, `action_space`, or `model_config`, you can specify these
        keys at the top level of the config, and the module spec will inherit the
        values from the top level config.

        Examples:

        .. code-block:: python

            config = {
                "modules": {
                    "module_1": SingleAgentRLModuleSpec(
                        module_class="RLModule1",
                        observation_space=gym.spaces.Box(...),
                        action_space=gym.spaces.Discrete(...),
                        model_config={hidden_dim: 256}
                    )
                    "module_2": SingleAgentRLModuleSpec(
                        module_class="RLModule2",
                        observation_space=gym.spaces.Box(...),
                    )
                },
                "action_space": gym.spaces.Box(...),
                "model_config": {hidden_dim: 32}
            }

            # This is equivalent to the following config:

            config = {
                "modules": {
                    "module_1": SingleAgentRLModuleSpec(
                        module_class="RLModule1",
                        observation_space=gym.spaces.Box(...),
                        action_space=gym.spaces.Discrete(...),
                        model_config={hidden_dim: 256}
                    )
                    "module_2": SingleAgentRLModuleSpec(
                        module_class="RLModule2",
                        observation_space=gym.spaces.Box(...),
                        action_space=gym.spaces.Box(...), # Inherited
                        model_config={hidden_dim: 32} # Inherited
                    }
                },
            }

        Args:
            config: A config dict that contains the module configs. See above for the
                format required.

            Returns:
                The MultiAgentRLModule.
        """

        module_configs: Dict[ModuleID, Any] = _get_module_configs(config)
        cls.__check_module_configs(module_configs)

        multiagent_module = cls()

        for module_id, module_spec in module_configs.items():
            # module_cls: RLModule = module_spec.pop("module_class")
            # module = module_cls.from_model_config(**module_spec)
            module = module_spec.build()
            multiagent_module.add_module(module_id, module)

        return multiagent_module

    @classmethod
    def __check_module_configs(cls, module_configs: Dict[ModuleID, Any]):
        """Checks the module configs for validity.

        The module_configs be a mapping from module_ids to SingleAgentRLModuleSpec
        objects.

        Args:
            module_configs: The module configs to check.

        Raises:
            ValueError: If the module configs are invalid.
        """
        for module_id, module_spec in module_configs.items():
            if not isinstance(module_spec, SingleAgentRLModuleSpec):
                raise ValueError(
                    f"Module {module_id} is not a SingleAgentRLModuleSpec object."
                )

    def keys(self) -> Iterator[ModuleID]:
        """Returns an iteratable of module ids."""
        return self._rl_modules.keys()

    @override(RLModule)
    def as_multi_agent(self) -> "MultiAgentRLModule":
        """Returns a multi-agent wrapper around this module.

        This method is overridden to avoid double wrapping.

        Returns:
            The instance itself.
        """
        return self

    def add_module(
        self,
        module_id: ModuleID,
        module: RLModule,
        *,
        override: bool = False,
    ) -> None:
        """Adds a module at run time to the multi-agent module.

        Args:
            module_id: The module ID to add. If the module ID already exists and
                override is False, an error is raised. If override is True, the module
                is replaced.
            module: The module to add.
            override: Whether to override the module if it already exists.

        Raises:
            ValueError: If the module ID already exists and override is False.
            Warnings are raised if the module id is not valid according to the logic of
            validate_policy_id().
        """
        validate_policy_id(module_id)
        if module_id in self._rl_modules and not override:
            raise ValueError(
                f"Module ID {module_id} already exists. If your intention is to "
                "override, set override=True."
            )
        self._rl_modules[module_id] = module

    def remove_module(
        self, module_id: ModuleID, *, raise_err_if_not_found: bool = True
    ) -> None:
        """Removes a module at run time from the multi-agent module.

        Args:
            module_id: The module ID to remove.
            raise_err_if_not_found: Whether to raise an error if the module ID is not
                found.
        Raises:
            ValueError: If the module ID does not exist and raise_err_if_not_found is
                True.
        """
        if raise_err_if_not_found:
            self._check_module_exists(module_id)
        del self._rl_modules[module_id]

    @override(RLModule)
    def make_distributed(self, dist_config: Mapping[str, Any] = None) -> None:
        # TODO (Avnish) Implement this.
        pass

    @override(RLModule)
    def is_distributed(self) -> bool:
        # TODO (Avnish) Implement this.
        return False

    def __getitem__(self, module_id: ModuleID) -> RLModule:
        """Returns the module with the given module ID.

        Args:
            module_id: The module ID to get.

        Returns:
            The module with the given module ID.
        """
        self._check_module_exists(module_id)
        return self._rl_modules[module_id]

    @override(RLModule)
    def output_specs_train(self) -> SpecDict:
        return self._get_specs_for_modules("output_specs_train")

    @override(RLModule)
    def output_specs_inference(self) -> SpecDict:
        return self._get_specs_for_modules("output_specs_inference")

    @override(RLModule)
    def output_specs_exploration(self) -> SpecDict:
        return self._get_specs_for_modules("output_specs_exploration")

    @override(RLModule)
    def input_specs_train(self) -> SpecDict:
        return self._get_specs_for_modules("input_specs_train")

    @override(RLModule)
    def input_specs_inference(self) -> SpecDict:
        return self._get_specs_for_modules("input_specs_inference")

    @override(RLModule)
    def input_specs_exploration(self) -> SpecDict:
        return self._get_specs_for_modules("input_specs_exploration")

    def _get_specs_for_modules(self, method_name: str) -> SpecDict:
        """Returns a ModelSpec from the given method_name for all modules."""
        return SpecDict(
            {
                module_id: getattr(module, method_name)()
                for module_id, module in self._rl_modules.items()
            }
        )

    @override(RLModule)
    def _forward_train(
        self, batch: MultiAgentBatch, **kwargs
    ) -> Union[Mapping[str, Any], Dict[ModuleID, Mapping[str, Any]]]:
        """Runs the forward_train pass.

        TODO(avnishn, kourosh): Review type hints for forward methods.

        Args:
            batch: The batch of multi-agent data (i.e. mapping from module ids to
                SampleBaches).

        Returns:
            The output of the forward_train pass the specified modules.
        """
        return self.__run_forward_pass("forward_train", batch, **kwargs)

    @override(RLModule)
    def _forward_inference(
        self, batch: MultiAgentBatch, **kwargs
    ) -> Union[Mapping[str, Any], Dict[ModuleID, Mapping[str, Any]]]:
        """Runs the forward_inference pass.

        TODO(avnishn, kourosh): Review type hints for forward methods.

        Args:
            batch: The batch of multi-agent data (i.e. mapping from module ids to
                SampleBaches).

        Returns:
            The output of the forward_inference pass the specified modules.
        """
        return self.__run_forward_pass("forward_inference", batch, **kwargs)

    @override(RLModule)
    def _forward_exploration(
        self, batch: MultiAgentBatch, **kwargs
    ) -> Union[Mapping[str, Any], Dict[ModuleID, Mapping[str, Any]]]:
        """Runs the forward_exploration pass.

        TODO(avnishn, kourosh): Review type hints for forward methods.

        Args:
            batch: The batch of multi-agent data (i.e. mapping from module ids to
                SampleBaches).

        Returns:
            The output of the forward_exploration pass the specified modules.
        """
        return self.__run_forward_pass("forward_exploration", batch, **kwargs)

    @override(RLModule)
    def get_state(self) -> Mapping[str, Any]:
        """Returns the state of the multi-agent module.

        The default implementation loops all modules and calls their get_state method.
        Override this method if you want to change the get_state behavior.

        Returns:
            A nested state dict with the first layer being the module ID and the second
            is the state of the module.
        """
        return {
            module_id: module.get_state()
            for module_id, module in self._rl_modules.items()
        }

    @override(RLModule)
    def set_state(self, state_dict: Mapping[str, Any]) -> None:
        """Sets the state dict of the multi-agent module.

        The default implementation is a mapping from independent module IDs to their
        corresponding RLModule state_dicts. Override this method to customize the
        state_dict for custom more advanced multi-agent use cases.

        Args:
            state_dict: The state dict to set.
        """
        for module_id, module in self._rl_modules.items():
            module.set_state(state_dict[module_id])

    def serialize(self) -> Mapping[str, Any]:
        """Return the serialized state of the module.

        NOTE: This method needs to be implemented in order to support
        checkpointing and fault tolerance.

        """
        return {
            "class": self.__class__,
            "rl_modules": {
                module_id: module.serialize()
                for module_id, module in self._rl_modules.items()
            },
        }

    @classmethod
    def deserialize(cls, state: Mapping[str, Any]) -> "MultiAgentRLModule":
        """Construct a module from a serialized state.

        Args:
            state: The serialized state of the module.
            The state should contain the keys "class", "kwargs", and "state".

            - "class" is the class of the RLModule to be constructed.
            - "rl_modules" is a dict mapping module ids of the RLModules to
                their serialized states. The serialized states can be obtained
                from `RLModule.serialize()`.

        NOTE: this state is typically obtained from `serialize()`.

        NOTE: This method needs to be implemented in order to support
            checkpointing and fault tolerance.

        Returns:
            A deserialized MultiAgentRLModule.
        """
        rl_modules = {}
        for module_id, module_state in state["rl_modules"].items():
            rl_modules[module_id] = RLModule.deserialize(module_state)
        return cls(rl_modules)

    def __repr__(self) -> str:
        return f"MARL({pprint.pformat(self._rl_modules)})"

    def __run_forward_pass(
        self,
        forward_fn_name: str,
        batch: NestedDict[Any],
        **kwargs,
    ) -> Dict[ModuleID, Mapping[str, Any]]:
        """This is a helper method that runs the forward pass for the given module.

        It uses forward_fn_name to get the forward pass method from the RLModule
        (e.g. forward_train vs. forward_exploration) and runs it on the given batch.

        Args:
            forward_fn_name: The name of the forward pass method to run.
            batch: The batch of multi-agent data (i.e. mapping from module ids to
                SampleBaches).
            **kwargs: Additional keyword arguments to pass to the forward function.

        Returns:
            The output of the forward pass the specified modules. The output is a
            mapping from module ID to the output of the forward pass.
        """

        module_ids = list(batch.shallow_keys())
        for module_id in module_ids:
            self._check_module_exists(module_id)

        outputs = {}
        for module_id in module_ids:
            rl_module = self._rl_modules[module_id]
            forward_fn = getattr(rl_module, forward_fn_name)
            outputs[module_id] = forward_fn(batch[module_id], **kwargs)

        return outputs

    def _check_module_exists(self, module_id: ModuleID) -> None:
        if module_id not in self._rl_modules:
            raise KeyError(
                f"Module with module_id {module_id} not found. "
                f"Available modules: {set(self.keys())}"
            )



import warnings
from typing import TYPE_CHECKING, Dict, Optional

from ray.air._internal.session import Session
from ray.air.checkpoint import Checkpoint

if TYPE_CHECKING:
    # avoid circular import
    from ray.data import DatasetIterator
    from ray.train._internal.session import _TrainSession
    from ray.tune.execution.placement_groups import PlacementGroupFactory


class _TrainSessionImpl(Session):
    """Session client that "per worker train loop" can interact with.

    Notice that each worker will automatically switch to its working
    directory on entering the train loop. This is to ensure that
    each worker can safely write to a local directory without racing
    and overwriting each other."""

    def __init__(self, session: "_TrainSession"):
        self._session = session

    def report(self, metrics: Dict, *, checkpoint: Optional[Checkpoint] = None) -> None:
        self._session.report(metrics, checkpoint)

    @property
    def loaded_checkpoint(self) -> Optional[Checkpoint]:
        ckpt = self._session.loaded_checkpoint
        if ckpt:
            # The new API should only interact with Checkpoint object.
            assert isinstance(ckpt, Checkpoint)
        return ckpt

    @property
    def experiment_name(self) -> str:
        return self._session.trial_info.experiment_name

    @property
    def trial_name(self) -> str:
        return self._session.trial_info.name

    @property
    def trial_id(self) -> str:
        return self._session.trial_info.id

    @property
    def trial_resources(self) -> "PlacementGroupFactory":
        return self._session.trial_info.resources

    @property
    def trial_dir(self) -> str:
        return self._session.trial_info.logdir

    @property
    def world_size(self) -> int:
        return self._session.world_size

    @property
    def world_rank(self) -> int:
        return self._session.world_rank

    @property
    def local_rank(self) -> int:
        return self._session.local_rank

    @property
    def local_world_size(self) -> int:
        return self._session.local_world_size

    @property
    def node_rank(self) -> int:
        return self._session.node_rank

    def get_dataset_shard(
        self,
        dataset_name: Optional[str] = None,
    ) -> Optional["DatasetIterator"]:
        shard = self._session.dataset_shard
        if shard is None:
            warnings.warn(
                "No dataset passed in. Returning None. Make sure to "
                "pass in a Ray Dataset to Trainer.run to use this "
                "function."
            )
        elif isinstance(shard, dict):
            if not dataset_name:
                raise RuntimeError(
                    "Multiple datasets were passed into ``Trainer``, "
                    "but no ``dataset_name`` is passed into "
                    "``get_dataset_shard``. Please specify which "
                    "dataset shard to retrieve."
                )
            return shard.get(dataset_name)
        return shard



"""Test the collective allreduice API on a distributed Ray cluster."""
import pytest
import logging

import cupy as cp

import ray
from ray.util.collective.types import ReduceOp
from ray.util.collective.tests.util import create_collective_multigpu_workers

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")


@pytest.mark.parametrize("group_name", ["default", "test", "123?34!"])
def test_allreduce_multigpu_different_name(
    ray_start_distributed_multigpu_2_nodes_4_gpus, group_name
):
    world_size = 2
    num_gpu_per_worker = 2
    actual_world_size = world_size * num_gpu_per_worker
    actors, _ = create_collective_multigpu_workers(
        num_workers=world_size, group_name=group_name
    )
    results = ray.get([a.do_allreduce_multigpu.remote(group_name) for a in actors])
    assert (results[0] == cp.ones((10,), dtype=cp.float32) * actual_world_size).all()
    assert (results[1] == cp.ones((10,), dtype=cp.float32) * actual_world_size).all()


@pytest.mark.parametrize("array_size", [2, 2**5, 2**10, 2**15, 2**20])
def test_allreduce_multigpu_different_array_size(
    ray_start_distributed_multigpu_2_nodes_4_gpus, array_size
):
    world_size = 2
    num_gpu_per_worker = 2
    actual_world_size = world_size * num_gpu_per_worker
    actors, _ = create_collective_multigpu_workers(world_size)
    ray.get([a.set_buffer.remote(array_size) for a in actors])
    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (
        results[0] == cp.ones((array_size,), dtype=cp.float32) * actual_world_size
    ).all()
    assert (
        results[1] == cp.ones((array_size,), dtype=cp.float32) * actual_world_size
    ).all()


def test_allreduce_multigpu_destroy(
    ray_start_distributed_multigpu_2_nodes_4_gpus, backend="nccl", group_name="default"
):
    world_size = 2
    num_gpu_per_worker = 2
    actual_world_size = world_size * num_gpu_per_worker
    actors, _ = create_collective_multigpu_workers(world_size)

    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (results[0] == cp.ones((10,), dtype=cp.float32) * actual_world_size).all()
    assert (results[1] == cp.ones((10,), dtype=cp.float32) * actual_world_size).all()

    # destroy the group and try do work, should fail
    ray.get([a.destroy_group.remote() for a in actors])
    with pytest.raises(RuntimeError):
        results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])

    # reinit the same group and all reduce
    ray.get(
        [
            actor.init_group.remote(world_size, i, backend, group_name)
            for i, actor in enumerate(actors)
        ]
    )
    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (
        results[0]
        == cp.ones((10,), dtype=cp.float32) * actual_world_size * actual_world_size
    ).all()
    assert (
        results[1]
        == cp.ones((10,), dtype=cp.float32) * actual_world_size * actual_world_size
    ).all()


def test_allreduce_multigpu_multiple_group(
    ray_start_distributed_multigpu_2_nodes_4_gpus, backend="nccl", num_groups=5
):
    world_size = 2
    num_gpu_per_worker = 2
    actual_world_size = world_size * num_gpu_per_worker
    actors, _ = create_collective_multigpu_workers(world_size)
    for group_name in range(1, num_groups):
        ray.get(
            [
                actor.init_group.remote(world_size, i, backend, str(group_name))
                for i, actor in enumerate(actors)
            ]
        )
    for i in range(num_groups):
        group_name = "default" if i == 0 else str(i)
        results = ray.get([a.do_allreduce_multigpu.remote(group_name) for a in actors])
        assert (
            results[0]
            == cp.ones((10,), dtype=cp.float32) * (actual_world_size ** (i + 1))
        ).all()


def test_allreduce_multigpu_different_op(ray_start_distributed_multigpu_2_nodes_4_gpus):
    world_size = 2
    actors, _ = create_collective_multigpu_workers(world_size)
    # check product
    ray.get(actors[0].set_buffer.remote([10], value0=2, value1=3))
    ray.get(actors[1].set_buffer.remote([10], value0=4, value1=5))
    results = ray.get(
        [a.do_allreduce_multigpu.remote(op=ReduceOp.PRODUCT) for a in actors]
    )
    assert (results[0] == cp.ones((10,), dtype=cp.float32) * 120).all()
    assert (results[1] == cp.ones((10,), dtype=cp.float32) * 120).all()

    # check min
    ray.get(actors[0].set_buffer.remote([10], value0=2, value1=3))
    ray.get(actors[1].set_buffer.remote([10], value0=4, value1=5))
    results = ray.get([a.do_allreduce_multigpu.remote(op=ReduceOp.MIN) for a in actors])
    assert (results[0] == cp.ones((10,), dtype=cp.float32) * 2).all()
    assert (results[1] == cp.ones((10,), dtype=cp.float32) * 2).all()

    # check max
    ray.get(actors[0].set_buffer.remote([10], value0=2, value1=3))
    ray.get(actors[1].set_buffer.remote([10], value0=4, value1=5))
    results = ray.get([a.do_allreduce_multigpu.remote(op=ReduceOp.MAX) for a in actors])
    assert (results[0] == cp.ones((10,), dtype=cp.float32) * 5).all()
    assert (results[1] == cp.ones((10,), dtype=cp.float32) * 5).all()


@pytest.mark.parametrize("dtype", [cp.uint8, cp.float16, cp.float32, cp.float64])
def test_allreduce_multigpu_different_dtype(
    ray_start_distributed_multigpu_2_nodes_4_gpus, dtype
):
    world_size = 2
    num_gpu_per_worker = 2
    actual_world_size = world_size * num_gpu_per_worker
    actors, _ = create_collective_multigpu_workers(world_size)
    ray.get([a.set_buffer.remote([10], dtype=dtype) for a in actors])
    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (results[0] == cp.ones((10,), dtype=dtype) * actual_world_size).all()
    assert (results[1] == cp.ones((10,), dtype=dtype) * actual_world_size).all()


def test_allreduce_torch_cupy(ray_start_distributed_multigpu_2_nodes_4_gpus):
    # import torch
    world_size = 2
    actual_world_size = 4
    actors, _ = create_collective_multigpu_workers(world_size)
    ray.get(actors[0].set_buffer.remote([10]))
    ray.get(
        actors[1].set_buffer.remote([10], tensor_type0="torch", tensor_type1="torch")
    )
    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (results[0] == cp.ones((10,)) * actual_world_size).all()

    ray.get(
        actors[0].set_buffer.remote([10], tensor_type0="cupy", tensor_type1="torch")
    )
    ray.get(
        actors[1].set_buffer.remote([10], tensor_type0="torch", tensor_type1="cupy")
    )
    results = ray.get([a.do_allreduce_multigpu.remote() for a in actors])
    assert (results[0] == cp.ones((10,)) * actual_world_size).all()



import os
import pickle
import re
import shutil
import tempfile
import unittest
from pathlib import Path
from typing import Any

import pytest

import ray
from ray.air._internal.remote_storage import _ensure_directory, delete_at_uri
from ray.air.checkpoint import _DICT_CHECKPOINT_ADDITIONAL_FILE_KEY, Checkpoint
from ray.air.constants import MAX_REPR_LENGTH, PREPROCESSOR_KEY
from ray.data import Preprocessor


class DummyPreprocessor(Preprocessor):
    def __init__(self, multiplier):
        self.multiplier = multiplier

    def transform_batch(self, df):
        return df * self.multiplier


def test_repr():
    checkpoint = Checkpoint(data_dict={"foo": "bar"})

    representation = repr(checkpoint)

    assert len(representation) < MAX_REPR_LENGTH
    pattern = re.compile("^Checkpoint\\((.*)\\)$")
    assert pattern.match(representation)


class StubCheckpoint(Checkpoint):

    _SERIALIZED_ATTRS = ("foo",)

    def __init__(self, *args, **kwargs):
        self.foo = None
        self.baz = None
        super().__init__(*args, **kwargs)


class OtherStubCheckpoint(Checkpoint):
    pass


class OtherStubCheckpointWithAttrs(Checkpoint):
    _SERIALIZED_ATTRS = StubCheckpoint._SERIALIZED_ATTRS


def test_from_checkpoint():
    checkpoint = Checkpoint.from_dict({"spam": "ham"})
    assert type(StubCheckpoint.from_checkpoint(checkpoint)) is StubCheckpoint

    # Check that attributes persist if same checkpoint type.
    checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
    checkpoint.foo = "bar"
    assert StubCheckpoint.from_checkpoint(checkpoint).foo == "bar"

    # Check that attributes persist if the new checkpoint
    # has them as well.
    # Check that attributes persist if same checkpoint type.
    checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
    checkpoint.foo = "bar"
    assert OtherStubCheckpointWithAttrs.from_checkpoint(checkpoint).foo == "bar"


class TestCheckpointTypeCasting:
    def test_dict(self):
        data = StubCheckpoint.from_dict({"foo": "bar"}).to_dict()
        assert isinstance(Checkpoint.from_dict(data), StubCheckpoint)

        data = Checkpoint.from_dict({"foo": "bar"}).to_dict()
        assert isinstance(StubCheckpoint.from_dict(data), StubCheckpoint)

        with pytest.raises(ValueError):
            data = OtherStubCheckpoint.from_dict({"foo": "bar"}).to_dict()
            StubCheckpoint.from_dict(data)

    def test_directory(self):
        path = StubCheckpoint.from_dict({"foo": "bar"}).to_directory()
        assert isinstance(Checkpoint.from_directory(path), StubCheckpoint)

        path = Checkpoint.from_dict({"foo": "bar"}).to_directory()
        assert isinstance(StubCheckpoint.from_directory(path), StubCheckpoint)

        with pytest.raises(ValueError):
            path = OtherStubCheckpoint.from_dict({"foo": "bar"}).to_directory()
            StubCheckpoint.from_directory(path)

    def test_uri(self):
        uri = StubCheckpoint.from_dict({"foo": "bar"}).to_uri("memory://1/")
        assert isinstance(Checkpoint.from_uri(uri), StubCheckpoint)

        uri = Checkpoint.from_dict({"foo": "bar"}).to_uri("memory://2/")
        assert isinstance(StubCheckpoint.from_uri(uri), StubCheckpoint)

        with pytest.raises(ValueError):
            uri = OtherStubCheckpoint.from_dict({"foo": "bar"}).to_uri("memory://3/")
            StubCheckpoint.from_uri(uri)

    def test_e2e(self):
        from ray.air import session
        from ray.air.config import ScalingConfig
        from ray.train.torch import TorchTrainer

        def train_loop_per_worker():
            checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
            session.report({}, checkpoint=checkpoint)

        trainer = TorchTrainer(
            train_loop_per_worker=train_loop_per_worker,
            scaling_config=ScalingConfig(num_workers=1),
        )
        results = trainer.fit()
        assert isinstance(results.checkpoint, StubCheckpoint)


class TestCheckpointSerializedAttrs:
    def test_dict(self):
        checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
        assert "foo" in checkpoint._SERIALIZED_ATTRS
        checkpoint.foo = "bar"

        recovered_checkpoint = StubCheckpoint.from_dict(checkpoint.to_dict())

        assert recovered_checkpoint.foo == "bar"

    def test_directory(self):
        checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
        assert "foo" in checkpoint._SERIALIZED_ATTRS
        checkpoint.foo = "bar"

        recovered_checkpoint = StubCheckpoint.from_directory(checkpoint.to_directory())

        assert recovered_checkpoint.foo == "bar"

    def test_directory_move_instead_of_copy(self):
        checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
        assert "foo" in checkpoint._SERIALIZED_ATTRS
        checkpoint.foo = "bar"

        path = checkpoint.to_directory()
        recovered_checkpoint = StubCheckpoint.from_directory(path)
        tmpdir = tempfile.mkdtemp()
        new_path = recovered_checkpoint._move_directory(tmpdir)
        new_recovered_checkpoint = StubCheckpoint.from_directory(new_path)

        assert recovered_checkpoint._local_path == tmpdir
        assert new_recovered_checkpoint.foo == "bar"
        assert not list(Path(path).glob("*"))

    def test_uri(self):
        checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
        assert "foo" in checkpoint._SERIALIZED_ATTRS
        checkpoint.foo = "bar"

        uri = checkpoint.to_uri("memory://bucket")
        recovered_checkpoint = StubCheckpoint.from_uri(uri)

        assert recovered_checkpoint.foo == "bar"

    def test_e2e(self):
        from ray.air import session
        from ray.air.config import ScalingConfig
        from ray.train.torch import TorchTrainer

        def train_loop_per_worker():
            checkpoint = StubCheckpoint.from_dict({"spam": "ham"})
            assert "foo" in checkpoint._SERIALIZED_ATTRS
            checkpoint.foo = "bar"
            session.report({}, checkpoint=checkpoint)

        trainer = TorchTrainer(
            train_loop_per_worker=train_loop_per_worker,
            scaling_config=ScalingConfig(num_workers=1),
        )
        results = trainer.fit()
        assert results.checkpoint.foo == "bar"


class CheckpointsConversionTest(unittest.TestCase):
    def setUp(self):
        self.tmpdir = os.path.realpath(tempfile.mkdtemp())
        self.tmpdir_pa = os.path.realpath(tempfile.mkdtemp())

        self.checkpoint_dict_data = {"metric": 5, "step": 4}
        self.checkpoint_dir_data = {"metric": 2, "step": 6}

        # We test two different in-memory filesystems as "cloud" providers,
        # one for fsspec and one for pyarrow.fs

        # fsspec URI
        self.cloud_uri = "memory:///cloud/bucket"
        # pyarrow URI
        self.cloud_uri_pa = "mock://cloud/bucket/"

        self.checkpoint_dir = os.path.join(self.tmpdir, "existing_checkpoint")
        os.mkdir(self.checkpoint_dir, 0o755)
        with open(os.path.join(self.checkpoint_dir, "test_data.pkl"), "wb") as fp:
            pickle.dump(self.checkpoint_dir_data, fp)

        self.old_dir = os.getcwd()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.old_dir)
        shutil.rmtree(self.tmpdir)
        shutil.rmtree(self.tmpdir_pa)

    def _prepare_dict_checkpoint(self) -> StubCheckpoint:
        # Create checkpoint from dict
        checkpoint = StubCheckpoint.from_dict(dict(self.checkpoint_dict_data))
        # The `foo` attribute should be serialized.
        self.assertTrue("foo" in StubCheckpoint._SERIALIZED_ATTRS)
        checkpoint.foo = "bar"
        # The `baz` attribute shouldn't be serialized.
        self.assertFalse("baz" in StubCheckpoint._SERIALIZED_ATTRS)
        checkpoint.baz = "qux"

        self.assertIsInstance(checkpoint, StubCheckpoint)
        self.assertTrue(checkpoint._data_dict)
        self.assertEqual(
            checkpoint._data_dict["metric"], self.checkpoint_dict_data["metric"]
        )

        return checkpoint

    def _assert_dict_checkpoint(self, checkpoint, check_state=True):
        # Convert into dict
        checkpoint_data = checkpoint.to_dict()
        self.assertIsInstance(checkpoint, StubCheckpoint)
        if check_state:
            self.assertEqual(checkpoint.foo, "bar")
            self.assertEqual(checkpoint.baz, None)
        checkpoint_data = {
            key: value
            for key, value in checkpoint_data.items()
            if not key.startswith("_")
        }
        self.assertDictEqual(checkpoint_data, self.checkpoint_dict_data)

    def test_dict_checkpoint_bytes(self):
        """Test conversion from dict to bytes checkpoint and back."""
        checkpoint = self._prepare_dict_checkpoint()

        # Convert into bytes checkpoint
        blob = checkpoint.to_bytes()
        self.assertIsInstance(blob, bytes)

        # Create from bytes
        checkpoint = Checkpoint.from_bytes(blob)
        self.assertTrue(checkpoint._data_dict)

        self._assert_dict_checkpoint(checkpoint)

    def test_dict_checkpoint_dict(self):
        """Test conversion from dict to dict checkpoint and back."""
        checkpoint = self._prepare_dict_checkpoint()

        # Convert into dict checkpoint
        data_dict = checkpoint.to_dict()
        self.assertIsInstance(data_dict, dict)

        # Create from dict
        checkpoint = Checkpoint.from_dict(data_dict)
        self.assertTrue(checkpoint._data_dict)

        self._assert_dict_checkpoint(checkpoint)

    def test_dict_checkpoint_fs(self):
        """Test conversion from dict to FS checkpoint and back."""
        checkpoint = self._prepare_dict_checkpoint()

        # Convert into fs checkpoint
        path = checkpoint.to_directory()
        self.assertIsInstance(path, str)

        # Create from path
        checkpoint = Checkpoint.from_directory(path)
        self.assertTrue(checkpoint._local_path)

        self._assert_dict_checkpoint(checkpoint)

    def test_dict_checkpoint_obj_store(self):
        """Test conversion from fs to obj store checkpoint and back."""
        if not ray.is_initialized():
            ray.init()

        checkpoint = self._prepare_dict_checkpoint()

        # Convert into dict checkpoint
        obj_ref = ray.put(checkpoint)
        self.assertIsInstance(obj_ref, ray.ObjectRef)

        # Create from dict
        checkpoint = ray.get(obj_ref)
        self._assert_dict_checkpoint(checkpoint, check_state=False)

    def test_dict_checkpoint_uri(self):
        """Test conversion from dict to cloud checkpoint and back."""
        checkpoint = self._prepare_dict_checkpoint()

        # Convert into dict checkpoint
        location = checkpoint.to_uri(self.cloud_uri)
        self.assertIsInstance(location, str)
        self.assertIn("memory://", location)

        # Create from dict
        checkpoint = Checkpoint.from_uri(location)
        self.assertTrue(checkpoint._uri)

        self._assert_dict_checkpoint(checkpoint)

    def test_dict_checkpoint_uri_pa(self):
        """Test conversion from dict to cloud checkpoint and back."""
        checkpoint = self._prepare_dict_checkpoint()

        # Clean up mock bucket
        delete_at_uri(self.cloud_uri_pa)
        _ensure_directory(self.cloud_uri_pa)

        # Convert into dict checkpoint
        location = checkpoint.to_uri(self.cloud_uri_pa)
        self.assertIsInstance(location, str)
        self.assertIn("mock://", location)

        # Create from dict
        checkpoint = Checkpoint.from_uri(location)
        self.assertTrue(checkpoint._uri)

        self._assert_dict_checkpoint(checkpoint)

    def _prepare_fs_checkpoint(self) -> StubCheckpoint:
        # Create checkpoint from fs
        checkpoint = StubCheckpoint.from_directory(self.checkpoint_dir)
        # The `foo` attribute should be serialized.
        self.assertIn("foo", StubCheckpoint._SERIALIZED_ATTRS)
        checkpoint.foo = "bar"
        # The `baz` attribute shouldn't be serialized.
        self.assertNotIn("baz", StubCheckpoint._SERIALIZED_ATTRS)
        checkpoint.baz = "qux"

        self.assertTrue(checkpoint._local_path, str)
        self.assertEqual(checkpoint._local_path, self.checkpoint_dir)

        return checkpoint

    def _assert_fs_checkpoint(self, checkpoint):
        # Convert back to directory
        local_dir = checkpoint.to_directory()

        with open(os.path.join(local_dir, "test_data.pkl"), "rb") as fp:
            local_data = pickle.load(fp)

        self.assertEqual(checkpoint.foo, "bar")
        self.assertEqual(checkpoint.baz, None)
        self.assertIsInstance(checkpoint, StubCheckpoint)
        self.assertDictEqual(local_data, self.checkpoint_dir_data)

    def test_fs_checkpoint_bytes(self):
        """Test conversion from fs to bytes checkpoint and back."""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into bytest checkpoint
        blob = checkpoint.to_bytes()
        self.assertIsInstance(blob, bytes)

        # Create from bytes
        checkpoint = Checkpoint.from_bytes(blob)
        self.assertTrue(checkpoint._data_dict)

        self._assert_fs_checkpoint(checkpoint)

    def test_fs_checkpoint_dict(self):
        """Test conversion from fs to dict checkpoint and back."""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into dict checkpoint
        data_dict = checkpoint.to_dict()
        self.assertIsInstance(data_dict, dict)

        # Create from dict
        checkpoint = Checkpoint.from_dict(data_dict)
        self.assertTrue(checkpoint._data_dict)

        self._assert_fs_checkpoint(checkpoint)

    def test_metadata(self):
        """Test conversion with metadata involved.

        a. from fs to dict checkpoint;
        b. drop some marker to dict checkpoint;
        c. convert back to fs checkpoint;
        d. convert back to dict checkpoint.

        Assert that the marker should still be there."""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into dict checkpoint
        data_dict = checkpoint.to_dict()
        self.assertIsInstance(data_dict, dict)

        data_dict["my_marker"] = "marked"

        # Create from dict
        checkpoint = Checkpoint.from_dict(data_dict)
        self.assertTrue(checkpoint._data_dict)

        self._assert_fs_checkpoint(checkpoint)

        # Convert back to dict
        data_dict_2 = Checkpoint.from_directory(checkpoint.to_directory()).to_dict()
        assert data_dict_2["my_marker"] == "marked"

    def test_fs_checkpoint_fs(self):
        """Test conversion from fs to fs checkpoint and back."""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into fs checkpoint
        path = checkpoint.to_directory()
        self.assertIsInstance(path, str)

        # Create from fs
        checkpoint = Checkpoint.from_directory(path)
        self.assertTrue(checkpoint._local_path)

        self._assert_fs_checkpoint(checkpoint)

    def test_fs_checkpoint_obj_store(self):
        """Test conversion from fs to obj store checkpoint and back."""
        if not ray.is_initialized():
            ray.init()

        checkpoint = self._prepare_fs_checkpoint()

        # Convert into obj ref checkpoint
        obj_ref = ray.put(checkpoint)

        # Create from object ref
        checkpoint = ray.get(obj_ref)
        self._assert_fs_checkpoint(checkpoint)

    def test_fs_checkpoint_uri(self):
        """Test conversion from fs to cloud checkpoint and back."""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into dict checkpoint
        location = checkpoint.to_uri(self.cloud_uri)
        self.assertIsInstance(location, str)
        self.assertIn("memory://", location)

        # Create from dict
        checkpoint = Checkpoint.from_uri(location)
        self.assertTrue(checkpoint._uri)

        self._assert_fs_checkpoint(checkpoint)

    def test_fs_checkpoint_uri_pa(self):
        """Test conversion from fs to cloud checkpoint and back."""
        checkpoint = self._prepare_fs_checkpoint()

        # Clean up mock bucket
        delete_at_uri(self.cloud_uri_pa)
        _ensure_directory(self.cloud_uri_pa)

        # Convert into dict checkpoint
        location = checkpoint.to_uri(self.cloud_uri_pa)
        self.assertIsInstance(location, str)
        self.assertIn("mock://", location)

        # Create from dict
        checkpoint = Checkpoint.from_uri(location)
        self.assertTrue(checkpoint._uri)

        self._assert_fs_checkpoint(checkpoint)

    def test_fs_delete_at_uri(self):
        """Test that clear bucket utility works"""
        checkpoint = self._prepare_fs_checkpoint()

        # Convert into dict checkpoint
        location = checkpoint.to_uri(self.cloud_uri)
        delete_at_uri(location)

        checkpoint = Checkpoint.from_uri(location)
        with self.assertRaises(FileNotFoundError):
            checkpoint.to_directory()

    def test_fs_cp_as_directory(self):
        checkpoint = self._prepare_fs_checkpoint()

        with checkpoint.as_directory() as checkpoint_dir:
            assert checkpoint._local_path == checkpoint_dir

        assert os.path.exists(checkpoint_dir)

    def test_dict_cp_as_directory(self):
        checkpoint = self._prepare_dict_checkpoint()

        with checkpoint.as_directory() as checkpoint_dir:
            assert os.path.exists(checkpoint_dir)

        assert not os.path.exists(checkpoint_dir)

    def test_obj_store_cp_as_directory(self):
        checkpoint = self._prepare_dict_checkpoint()

        # Convert into obj ref checkpoint
        obj_ref = ray.put(checkpoint)

        # Create from object ref
        checkpoint = ray.get(obj_ref)

        with checkpoint.as_directory() as checkpoint_dir:
            assert os.path.exists(checkpoint_dir)
            assert Path(checkpoint_dir).stem.endswith(checkpoint._uuid.hex)

        assert not os.path.exists(checkpoint_dir)

    def test_dict_checkpoint_additional_files(self):
        checkpoint = self._prepare_dict_checkpoint()

        # Convert to directory
        checkpoint_dir = checkpoint.to_directory()

        # Add file into checkpoint directory
        with open(os.path.join(checkpoint_dir, "additional_file.txt"), "w") as f:
            f.write("Additional data\n")
        os.mkdir(os.path.join(checkpoint_dir, "subdir"))
        with open(os.path.join(checkpoint_dir, "subdir", "another.txt"), "w") as f:
            f.write("Another additional file\n")

        # Create new checkpoint object
        checkpoint = Checkpoint.from_directory(checkpoint_dir)

        new_dir = checkpoint.to_directory()

        assert os.path.exists(os.path.join(new_dir, "additional_file.txt"))
        with open(os.path.join(new_dir, "additional_file.txt"), "r") as f:
            assert f.read() == "Additional data\n"

        assert os.path.exists(os.path.join(new_dir, "subdir", "another.txt"))
        with open(os.path.join(new_dir, "subdir", "another.txt"), "r") as f:
            assert f.read() == "Another additional file\n"

        checkpoint_dict = checkpoint.to_dict()
        for k, v in self.checkpoint_dict_data.items():
            assert checkpoint_dict[k] == v

        assert _DICT_CHECKPOINT_ADDITIONAL_FILE_KEY in checkpoint_dict

        # Add another field
        checkpoint_dict["new_field"] = "Data"

        another_dict = Checkpoint.from_directory(
            Checkpoint.from_dict(checkpoint_dict).to_directory()
        ).to_dict()
        assert _DICT_CHECKPOINT_ADDITIONAL_FILE_KEY in another_dict
        assert another_dict["new_field"] == "Data"

    def test_fs_checkpoint_additional_fields(self):
        checkpoint = self._prepare_fs_checkpoint()

        # Convert to dict
        checkpoint_dict = checkpoint.to_dict()

        # Add field to dict
        checkpoint_dict["additional_field"] = "data"

        # Create new checkpoint object
        checkpoint = Checkpoint.from_dict(checkpoint_dict)

        # Turn into FS
        checkpoint_dir = checkpoint.to_directory()

        assert os.path.exists(os.path.join(checkpoint_dir, "test_data.pkl"))
        assert os.path.exists(os.path.join(checkpoint_dir, "additional_field.meta.pkl"))

        # Add new file
        with open(os.path.join(checkpoint_dir, "even_more.txt"), "w") as f:
            f.write("More\n")

        # Turn into dict
        new_dict = Checkpoint.from_directory(checkpoint_dir).to_dict()

        assert new_dict["additional_field"] == "data"

        # Turn into fs
        new_dir = Checkpoint.from_dict(new_dict).to_directory()

        assert os.path.exists(os.path.join(new_dir, "test_data.pkl"))
        assert os.path.exists(os.path.join(new_dir, "additional_field.meta.pkl"))
        assert os.path.exists(os.path.join(new_dir, "even_more.txt"))


class CheckpointsSerdeTest(unittest.TestCase):
    def setUp(self) -> None:
        if not ray.is_initialized():
            ray.init()

    def _testCheckpointSerde(
        self, checkpoint: Checkpoint, expected_type: str, expected_data: Any
    ):
        @ray.remote
        def assert_checkpoint_content(cp: Checkpoint):
            type_, data = cp.get_internal_representation()
            assert type_ == expected_type
            assert data == expected_data

            return True

        self.assertTrue(ray.get(assert_checkpoint_content.remote(checkpoint)))

    def testUriCheckpointSerde(self):
        # URI checkpoints keep the same internal representation, pointing to
        # a remote location

        checkpoint = Checkpoint.from_uri("memory:///some/bucket")

        self._testCheckpointSerde(checkpoint, *checkpoint.get_internal_representation())

    def testDataCheckpointSerde(self):
        # Data checkpoints keep the same internal representation, including
        # their data.

        checkpoint = Checkpoint.from_dict({"checkpoint_data": 5})

        self._testCheckpointSerde(checkpoint, *checkpoint.get_internal_representation())

    def testLocalCheckpointSerde(self):
        # Local checkpoints are converted to bytes on serialization. Currently
        # this is a pickled dict, so we compare with a dict checkpoint.
        source_checkpoint = Checkpoint.from_dict({"checkpoint_data": 5})
        with source_checkpoint.as_directory() as tmpdir:
            checkpoint = Checkpoint.from_directory(tmpdir)
            self._testCheckpointSerde(
                checkpoint, *source_checkpoint.get_internal_representation()
            )

    def testBytesCheckpointSerde(self):
        # Bytes checkpoints are just dict checkpoints constructed
        # from pickled data, so we compare with the source dict checkpoint.
        source_checkpoint = Checkpoint.from_dict({"checkpoint_data": 5})
        blob = source_checkpoint.to_bytes()
        checkpoint = Checkpoint.from_bytes(blob)

        self._testCheckpointSerde(
            checkpoint, *source_checkpoint.get_internal_representation()
        )

    def testObjRefCheckpointSerde(self):
        # Obj ref checkpoints are dict checkpoints put into the Ray object
        # store, but they have their own data representation (the obj ref).
        # We thus compare with the actual obj ref checkpoint.
        source_checkpoint = Checkpoint.from_dict({"checkpoint_data": 5})
        obj_ref = ray.put(source_checkpoint)
        checkpoint = ray.get(obj_ref)

        self._testCheckpointSerde(checkpoint, *checkpoint.get_internal_representation())


class PreprocessorCheckpointTest(unittest.TestCase):
    def testDictCheckpointWithoutPreprocessor(self):
        data = {"metric": 5}
        checkpoint = Checkpoint.from_dict(data)
        preprocessor = checkpoint.get_preprocessor()
        assert preprocessor is None

    def testDictCheckpointWithPreprocessor(self):
        preprocessor = DummyPreprocessor(1)
        data = {"metric": 5, PREPROCESSOR_KEY: preprocessor}
        checkpoint = Checkpoint.from_dict(data)
        preprocessor = checkpoint.get_preprocessor()
        assert preprocessor.multiplier == 1

    def testDictCheckpointWithPreprocessorAsDir(self):
        preprocessor = DummyPreprocessor(1)
        data = {"metric": 5, PREPROCESSOR_KEY: preprocessor}
        checkpoint = Checkpoint.from_dict(data)
        checkpoint_path = checkpoint.to_directory()
        checkpoint = Checkpoint.from_directory(checkpoint_path)
        preprocessor = checkpoint.get_preprocessor()
        assert preprocessor.multiplier == 1

    def testDirCheckpointWithoutPreprocessor(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            data = {"metric": 5}
            checkpoint_dir = os.path.join(tmpdir, "existing_checkpoint")
            os.mkdir(checkpoint_dir, 0o755)
            with open(os.path.join(checkpoint_dir, "test_data.pkl"), "wb") as fp:
                pickle.dump(data, fp)
            checkpoint = Checkpoint.from_directory(checkpoint_dir)
            preprocessor = checkpoint.get_preprocessor()
            assert preprocessor is None

    def testDirCheckpointWithPreprocessor(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            preprocessor = DummyPreprocessor(1)
            data = {"metric": 5}
            checkpoint_dir = os.path.join(tmpdir, "existing_checkpoint")
            os.mkdir(checkpoint_dir, 0o755)
            with open(os.path.join(checkpoint_dir, "test_data.pkl"), "wb") as fp:
                pickle.dump(data, fp)
            with open(os.path.join(checkpoint_dir, PREPROCESSOR_KEY), "wb") as fp:
                pickle.dump(preprocessor, fp)
            checkpoint = Checkpoint.from_directory(checkpoint_dir)
            preprocessor = checkpoint.get_preprocessor()
            assert preprocessor.multiplier == 1

    def testDirCheckpointWithPreprocessorAsDict(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            preprocessor = DummyPreprocessor(1)
            data = {"metric": 5}
            checkpoint_dir = os.path.join(tmpdir, "existing_checkpoint")
            os.mkdir(checkpoint_dir, 0o755)
            with open(os.path.join(checkpoint_dir, "test_data.pkl"), "wb") as fp:
                pickle.dump(data, fp)
            with open(os.path.join(checkpoint_dir, PREPROCESSOR_KEY), "wb") as fp:
                pickle.dump(preprocessor, fp)
            checkpoint = Checkpoint.from_directory(checkpoint_dir)
            checkpoint_dict = checkpoint.to_dict()
            checkpoint = checkpoint.from_dict(checkpoint_dict)
            preprocessor = checkpoint.get_preprocessor()
            assert preprocessor.multiplier == 1

    def testDictCheckpointSetPreprocessor(self):
        preprocessor = DummyPreprocessor(1)
        data = {"metric": 5}
        checkpoint = Checkpoint.from_dict(data)
        checkpoint.set_preprocessor(preprocessor)
        preprocessor = checkpoint.get_preprocessor()
        assert preprocessor.multiplier == 1

    def testDictCheckpointSetPreprocessorAsDir(self):
        preprocessor = DummyPreprocessor(1)
        data = {"metric": 5}
        checkpoint = Checkpoint.from_dict(data)
        checkpoint.set_preprocessor(preprocessor)
        checkpoint_path = checkpoint.to_directory()
        checkpoint = Checkpoint.from_directory(checkpoint_path)
        preprocessor = checkpoint.get_preprocessor()
        assert preprocessor.multiplier == 1

    def testDirCheckpointSetPreprocessor(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            preprocessor = DummyPreprocessor(1)
            data = {"metric": 5}
            checkpoint_dir = os.path.join(tmpdir, "existing_checkpoint")
            os.mkdir(checkpoint_dir, 0o755)
            with open(os.path.join(checkpoint_dir, "test_data.pkl"), "wb") as fp:
                pickle.dump(data, fp)
            checkpoint = Checkpoint.from_directory(checkpoint_dir)
            checkpoint.set_preprocessor(preprocessor)
            preprocessor = checkpoint.get_preprocessor()
            assert preprocessor.multiplier == 1

    def testDirCheckpointSetPreprocessorAsDict(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            preprocessor = DummyPreprocessor(1)
            data = {"metric": 5}
            checkpoint_dir = os.path.join(tmpdir, "existing_checkpoint")
            os.mkdir(checkpoint_dir, 0o755)
            with open(os.path.join(checkpoint_dir, "test_data.pkl"), "wb") as fp:
                pickle.dump(data, fp)
            checkpoint = Checkpoint.from_directory(checkpoint_dir)
            checkpoint.set_preprocessor(preprocessor)
            checkpoint_dict = checkpoint.to_dict()
            checkpoint = checkpoint.from_dict(checkpoint_dict)
            preprocessor = checkpoint.get_preprocessor()
            assert preprocessor.multiplier == 1

    def testObjectRefCheckpointSetPreprocessor(self):
        ckpt = Checkpoint.from_dict({"x": 1})
        ckpt = ray.get(ray.put(ckpt))
        preprocessor = DummyPreprocessor(1)
        ckpt.set_preprocessor(preprocessor)

        assert ckpt.get_preprocessor() == preprocessor

    def testAttrPath(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            checkpoint = Checkpoint.from_directory(tmpdir)
            with self.assertRaises(TypeError):
                os.path.exists(checkpoint)

    def testCheckpointUri(self):
        orig_checkpoint = Checkpoint.from_dict({"data": 2})

        self.assertEqual(orig_checkpoint.uri, None)

        # file:// URI
        with tempfile.TemporaryDirectory() as tmpdir:
            checkpoint = Checkpoint.from_directory(orig_checkpoint.to_directory(tmpdir))
            self.assertEqual(checkpoint.uri, "file://" + tmpdir)

        # cloud URI
        checkpoint = Checkpoint.from_uri(
            orig_checkpoint.to_uri("memory://some/location")
        )
        self.assertEqual(checkpoint.uri, "memory://some/location")


if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(["-v", __file__]))



import math
from typing import Callable, Iterable, List, Optional, Union

import numpy as np

from ray.data._internal.delegating_block_builder import DelegatingBlockBuilder
from ray.data._internal.execution.interfaces import TaskContext
from ray.data._internal.push_based_shuffle import PushBasedShufflePlan
from ray.data._internal.shuffle import ShuffleOp, SimpleShufflePlan
from ray.data.block import Block, BlockAccessor, BlockExecStats, BlockMetadata


class _ShufflePartitionOp(ShuffleOp):
    """
    Operator used for `random_shuffle` and `repartition` transforms.
    """

    def __init__(
        self,
        block_udf=None,
        random_shuffle: bool = False,
        random_seed: Optional[int] = None,
    ):
        super().__init__(
            map_args=[block_udf, random_shuffle, random_seed],
            reduce_args=[random_shuffle, random_seed],
        )

    @staticmethod
    def map(
        idx: int,
        block: Block,
        output_num_blocks: int,
        block_udf: Optional[Callable[[Block], Iterable[Block]]],
        random_shuffle: bool,
        random_seed: Optional[int],
    ) -> List[Union[BlockMetadata, Block]]:
        stats = BlockExecStats.builder()
        if block_udf:
            ctx = TaskContext(task_idx=idx)
            # TODO(ekl) note that this effectively disables block splitting.
            blocks = list(block_udf([block], ctx))
            if len(blocks) > 1:
                builder = BlockAccessor.for_block(blocks[0]).builder()
                for b in blocks:
                    builder.add_block(b)
                block = builder.build()
            else:
                block = blocks[0]
        block = BlockAccessor.for_block(block)

        # Randomize the distribution of records to blocks.
        if random_shuffle:
            seed_i = random_seed + idx if random_seed is not None else None
            block = block.random_shuffle(seed_i)
            block = BlockAccessor.for_block(block)

        slice_sz = max(1, math.ceil(block.num_rows() / output_num_blocks))
        slices = []
        for i in range(output_num_blocks):
            slices.append(block.slice(i * slice_sz, (i + 1) * slice_sz))

        # Randomize the distribution order of the blocks (this prevents empty
        # outputs when input blocks are very small).
        if random_shuffle:
            random = np.random.RandomState(seed_i)
            random.shuffle(slices)

        num_rows = sum(BlockAccessor.for_block(s).num_rows() for s in slices)
        assert num_rows == block.num_rows(), (num_rows, block.num_rows())
        metadata = block.get_metadata(input_files=None, exec_stats=stats.build())
        return slices + [metadata]

    @staticmethod
    def reduce(
        random_shuffle: bool,
        random_seed: Optional[int],
        *mapper_outputs: List[Block],
        partial_reduce: bool = False,
    ) -> (Block, BlockMetadata):
        stats = BlockExecStats.builder()
        builder = DelegatingBlockBuilder()
        for block in mapper_outputs:
            builder.add_block(block)
        new_block = builder.build()
        accessor = BlockAccessor.for_block(new_block)
        if random_shuffle:
            new_block = accessor.random_shuffle(
                random_seed if random_seed is not None else None
            )
            accessor = BlockAccessor.for_block(new_block)
        new_metadata = BlockMetadata(
            num_rows=accessor.num_rows(),
            size_bytes=accessor.size_bytes(),
            schema=accessor.schema(),
            input_files=None,
            exec_stats=stats.build(),
        )
        return new_block, new_metadata


class SimpleShufflePartitionOp(_ShufflePartitionOp, SimpleShufflePlan):
    pass


class PushBasedShufflePartitionOp(_ShufflePartitionOp, PushBasedShufflePlan):
    pass



"""
This module is higher-level abstraction of storage directly used by
workflows.
"""

import json
import logging
import os
import time
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import ray
from ray import cloudpickle
from ray._private import storage
from ray.types import ObjectRef
from ray.workflow.common import (
    TaskID,
    WorkflowStatus,
    WorkflowTaskRuntimeOptions,
)
from ray.workflow.exceptions import WorkflowNotFoundError
from ray.workflow import workflow_context
from ray.workflow import serialization
from ray.workflow import serialization_context
from ray.workflow.workflow_state import WorkflowExecutionState
from ray.workflow.storage import DataLoadError, DataSaveError, KeyNotFoundError

logger = logging.getLogger(__name__)

ArgsType = Tuple[List[Any], Dict[str, Any]]  # args and kwargs

# constants used for keys
WORKFLOW_ROOT = "workflows"  # The workflow root directory under global Ray storage.
OBJECTS_DIR = "objects"
STEPS_DIR = "tasks"
STEP_INPUTS_METADATA = "inputs.json"
STEP_USER_METADATA = "user_task_metadata.json"
STEP_PRERUN_METADATA = "pre_task_metadata.json"
STEP_POSTRUN_METADATA = "post_task_metadata.json"
STEP_OUTPUTS_METADATA = "outputs.json"
STEP_ARGS = "args.pkl"
STEP_OUTPUT = "output.pkl"
STEP_EXCEPTION = "exception.pkl"
STEP_FUNC_BODY = "func_body.pkl"
CLASS_BODY = "class_body.pkl"
WORKFLOW_META = "workflow_meta.json"
WORKFLOW_USER_METADATA = "user_run_metadata.json"
WORKFLOW_PRERUN_METADATA = "pre_run_metadata.json"
WORKFLOW_POSTRUN_METADATA = "post_run_metadata.json"
WORKFLOW_PROGRESS = "progress.json"
WORKFLOW_STATUS_DIR = "__status__"
WORKFLOW_STATUS_DIRTY_DIR = "dirty"
# Without this counter, we're going to scan all tasks to get the number of
# tasks with a given name. This can be very expensive if there are too
# many duplicates.
DUPLICATE_NAME_COUNTER = "duplicate_name_counter"


@dataclass
class TaskInspectResult:
    # The task output checkpoint exists and valid. If this field
    # is set, we do not set all other fields below.
    output_object_valid: bool = False
    # The ID of the task that could contain the output checkpoint of this
    # task. If this field is set, we do not set all other fields below.
    output_task_id: Optional[TaskID] = None
    # The task input arguments checkpoint exists and valid.
    args_valid: bool = False
    # The task function body checkpoint exists and valid.
    func_body_valid: bool = False
    # The dynamically referenced workflows in the input of the workflow.
    workflow_refs: Optional[List[str]] = None
    # The options of the workflow task.
    task_options: Optional[WorkflowTaskRuntimeOptions] = None
    # task throw exception
    task_raised_exception: bool = False

    def is_recoverable(self) -> bool:
        return (
            self.output_object_valid
            or self.output_task_id
            or (
                self.args_valid
                and self.workflow_refs is not None
                and self.func_body_valid
            )
        )


class WorkflowIndexingStorage:
    """Access and maintenance the indexing of workflow status.

    It runs a protocol that guarantees we can recover from any interrupted
    status updating. This protocol is **not thread-safe** for updating the
    status of the same workflow, currently it is executed by workflow management
    actor with a single thread.

    Here is how the protocol works:

    Update the status of a workflow
    1. Load workflow status from workflow data. If it is the same as the new status,
       return.
    2. Check if the workflow status updating is dirty. If it is, fix the
       workflow status; otherwise, mark the workflow status updating dirty.
    3. Update status in the workflow metadata.
    4. Insert the workflow ID key in the status indexing directory of the new status.
    5. Delete the workflow ID key in the status indexing directory of
       the previous status.
    6. Remove the workflow status updating dirty mark.

    Load a status of a workflow
    1. Read the status of the workflow from the workflow metadata.
    2. Return the status.

    List the status of all workflows
    1. Get status of all workflows by listing workflow ID keys in each workflow
       status indexing directory.
    2. List all workflows with dirty updating status. Get their status from
       workflow data. Override the status of the corresponding workflow.
    3. Return all the status.
    """

    def __init__(self):
        self._storage = storage.get_client(WORKFLOW_ROOT)

    def update_workflow_status(self, workflow_id: str, status: WorkflowStatus):
        """Update the status of the workflow.
        Try fixing indexing if workflow status updating was marked dirty.

        This method is NOT thread-safe. It is handled by the workflow management actor.
        """
        prev_status = self.load_workflow_status(workflow_id)
        if prev_status != status:
            # Try fixing indexing if workflow status updating was marked dirty.
            if (
                self._storage.get_info(self._key_workflow_status_dirty(workflow_id))
                is not None
            ):
                # This means the previous status update failed. Fix it.
                self._storage.put(
                    self._key_workflow_with_status(workflow_id, prev_status), b""
                )
                for s in WorkflowStatus:
                    if s != prev_status:
                        self._storage.delete(
                            self._key_workflow_with_status(workflow_id, s)
                        )
            else:
                self._storage.put(self._key_workflow_status_dirty(workflow_id), b"")
            # Transactional update of workflow status
            self._storage.put(
                self._key_workflow_metadata(workflow_id),
                json.dumps({"status": status.value}).encode(),
            )
            self._storage.put(self._key_workflow_with_status(workflow_id, status), b"")
            if prev_status is not WorkflowStatus.NONE:
                self._storage.delete(
                    self._key_workflow_with_status(workflow_id, prev_status)
                )
            self._storage.delete(self._key_workflow_status_dirty(workflow_id))

    def load_workflow_status(self, workflow_id: str):
        """Load the committed workflow status."""
        raw_data = self._storage.get(self._key_workflow_metadata(workflow_id))
        if raw_data is not None:
            metadata = json.loads(raw_data)
            return WorkflowStatus(metadata["status"])
        return WorkflowStatus.NONE

    def list_workflow(
        self, status_filter: Optional[Set[WorkflowStatus]] = None
    ) -> List[Tuple[str, WorkflowStatus]]:
        """List workflow status. Override status of the workflows whose status updating
        were marked dirty with the workflow status from workflow metadata.

        Args:
            status_filter: If given, only returns workflow with that status. This can
                be a single status or set of statuses.
        """
        if status_filter is None:
            status_filter = set(WorkflowStatus)
            status_filter.discard(WorkflowStatus.NONE)
        elif not isinstance(status_filter, set):
            raise TypeError("'status_filter' should either be 'None' or a set.")
        elif WorkflowStatus.NONE in status_filter:
            raise ValueError("'WorkflowStatus.NONE' is not a valid filter value.")

        results = {}
        for status in status_filter:
            try:
                # empty string points the key to the dir
                for p in self._storage.list(self._key_workflow_with_status("", status)):
                    workflow_id = p.base_name
                    results[workflow_id] = status
            except FileNotFoundError:
                pass
        # Get "correct" status of workflows
        try:
            for p in self._storage.list(self._key_workflow_status_dirty("")):
                workflow_id = p.base_name
                # overwrite status
                results.pop(workflow_id, None)
                status = self.load_workflow_status(workflow_id)
                if status in status_filter:
                    results[workflow_id] = status
        except FileNotFoundError:
            pass
        return list(results.items())

    def delete_workflow_status(self, workflow_id: str):
        """Delete status indexing for the workflow."""
        for status in WorkflowStatus:
            self._storage.delete(self._key_workflow_with_status(workflow_id, status))
        self._storage.delete(self._key_workflow_status_dirty(workflow_id))

    def _key_workflow_with_status(self, workflow_id: str, status: WorkflowStatus):
        """A key whose existence marks the status of the workflow."""
        return os.path.join(WORKFLOW_STATUS_DIR, status.value, workflow_id)

    def _key_workflow_status_dirty(self, workflow_id: str):
        """A key marks the workflow status dirty, because it is under change."""
        return os.path.join(WORKFLOW_STATUS_DIR, WORKFLOW_STATUS_DIRTY_DIR, workflow_id)

    def _key_workflow_metadata(self, workflow_id: str):
        return os.path.join(workflow_id, WORKFLOW_META)


class WorkflowStorage:
    """Access workflow in storage. This is a higher-level abstraction,
    which does not care about the underlining storage implementation."""

    def __init__(self, workflow_id: str):
        self._storage = storage.get_client(os.path.join(WORKFLOW_ROOT, workflow_id))
        self._status_storage = WorkflowIndexingStorage()
        self._workflow_id = workflow_id

    def load_task_output(self, task_id: TaskID) -> Any:
        """Load the output of the workflow task from checkpoint.

        Args:
            task_id: ID of the workflow task.

        Returns:
            Output of the workflow task.
        """

        tasks = [
            self._get(self._key_task_output(task_id), no_exception=True),
            self._get(self._key_task_exception(task_id), no_exception=True),
        ]
        (output_ret, output_err), (exception_ret, exception_err) = tasks
        # When we have output, always return output first
        if output_err is None:
            return output_ret

        # When we don't have output, check exception
        if exception_err is None:
            raise exception_ret

        # In this case, there is no such task
        raise output_err

    def save_workflow_execution_state(
        self, creator_task_id: TaskID, state: WorkflowExecutionState
    ) -> None:
        """Save a workflow execution state.
        Typically, the state is translated from a Ray DAG.

        Args:
            creator_task_id: The ID of the task that creates the state.
            state: The state converted from the DAG.
        """
        assert creator_task_id != state.output_task_id

        for task_id, task in state.tasks.items():
            # TODO (Alex): Handle the json case better?
            metadata = {
                **task.to_dict(),
                "workflow_refs": state.upstream_dependencies[task_id],
            }
            self._put(self._key_task_input_metadata(task_id), metadata, True)
            # TODO(suquark): The task user metadata duplicates.
            self._put(
                self._key_task_user_metadata(task_id),
                task.user_metadata,
                True,
            )
            workflow_id = self._workflow_id
            serialization.dump_to_storage(
                self._key_task_function_body(task_id),
                task.func_body,
                workflow_id,
                self,
            )
            with serialization_context.workflow_args_keeping_context():
                # TODO(suquark): in the future we should write to storage directly
                # with plasma store object in memory.
                args_obj = ray.get(state.task_input_args[task_id])
            serialization.dump_to_storage(
                self._key_task_args(task_id),
                args_obj,
                workflow_id,
                self,
            )

        # Finally, point to the output ID of the DAG. The DAG is a continuation
        # of the creator task.
        self._put(
            self._key_task_output_metadata(creator_task_id),
            {"output_task_id": state.output_task_id},
            True,
        )

    def save_task_output(
        self,
        task_id: TaskID,
        ret: Any,
        *,
        exception: Optional[Exception],
    ) -> None:
        """When a workflow task returns,
        1. If the returned object is a workflow, this means we are a nested
           workflow. We save the output metadata that points to the workflow.
        2. Otherwise, checkpoint the output.

        Args:
            task_id: The ID of the workflow task. If it is an empty string,
                it means we are in the workflow job driver process.
            ret: The returned object from a workflow task.
            exception: This task should throw exception.
        """
        if exception is None:
            # This workflow task returns a object.
            ret = ray.get(ret) if isinstance(ret, ray.ObjectRef) else ret
            serialization.dump_to_storage(
                self._key_task_output(task_id),
                ret,
                self._workflow_id,
                storage=self,
            )
            # tasks.append(self._put(self._key_task_output(task_id), ret))
            # TODO (yic): Delete exception file
        else:
            assert ret is None
            serialization.dump_to_storage(
                self._key_task_exception(task_id),
                exception,
                self._workflow_id,
                storage=self,
            )
            # tasks.append(
            #     self._put(self._key_task_exception(task_id), exception))

        # Finish checkpointing.
        # TODO(suquark): batching all tasks above.

    def load_task_func_body(self, task_id: TaskID) -> Callable:
        """Load the function body of the workflow task.

        Args:
            task_id: ID of the workflow task.

        Returns:
            A callable function.
        """
        return self._get(self._key_task_function_body(task_id))

    def gen_task_id(self, task_name: str) -> int:
        def _gen_task_id():
            key = self._key_num_tasks_with_name(task_name)
            try:
                val = self._get(key, True)
                self._put(key, val + 1, True)
                return val + 1
            except KeyNotFoundError:
                self._put(key, 0, True)
                return 0

        return _gen_task_id()

    def load_task_args(self, task_id: TaskID) -> ray.ObjectRef:
        """Load the input arguments of the workflow task. This must be
        done under a serialization context, otherwise the arguments would
        not be reconstructed successfully.

        Args:
            task_id: ID of the workflow task.

        Returns:
            An object ref of the input args.
        """
        with serialization_context.workflow_args_keeping_context():
            x = self._get(self._key_task_args(task_id))
        return ray.put(x)

    def save_object_ref(self, obj_ref: ray.ObjectRef) -> None:
        """Save the object ref.

        Args:
            obj_ref: The object reference

        Returns:
            None
        """
        return self._save_object_ref(obj_ref)

    def load_object_ref(self, object_id: str) -> ray.ObjectRef:
        """Load the input object ref.

        Args:
            object_id: The hex ObjectID.

        Returns:
            The object ref.
        """

        def _load_obj_ref() -> ray.ObjectRef:
            data = self._get(self._key_obj_id(object_id))
            ref = _put_obj_ref.remote((data,))
            return ref

        return _load_obj_ref()

    def update_continuation_output_link(
        self, continuation_root_id: TaskID, latest_continuation_task_id: TaskID
    ) -> None:
        """Update the link of the continuation output. The link points
        to the ID of the latest finished continuation task.

        Args:
            continuation_root_id: The ID of the task that returns all later
                continuations.
            latest_continuation_task_id: The ID of the latest finished
                continuation task.
        """
        try:
            metadata = self._get(
                self._key_task_output_metadata(continuation_root_id), True
            )
        except KeyNotFoundError:
            # This is because we skipped checkpointing of the
            # task [id=continuation_root_id]. Return a dummy
            # metadata instead.
            metadata = {}
        if latest_continuation_task_id != metadata.get(
            "output_task_id"
        ) and latest_continuation_task_id != metadata.get("dynamic_output_task_id"):
            metadata["dynamic_output_task_id"] = latest_continuation_task_id
            self._put(
                self._key_task_output_metadata(continuation_root_id), metadata, True
            )

    def _locate_output_task_id(self, task_id: TaskID) -> str:
        metadata = self._get(self._key_task_output_metadata(task_id), True)
        return metadata.get("dynamic_output_task_id") or metadata["output_task_id"]

    def get_entrypoint_task_id(self) -> TaskID:
        """Load the entrypoint task ID of the workflow.

        Returns:
            The ID of the entrypoint task.
        """
        # empty TaskID represents the workflow driver
        try:
            return self._locate_output_task_id("")
        except Exception as e:
            raise ValueError(
                "Fail to get entrypoint task ID from workflow"
                f"[id={self._workflow_id}]"
            ) from e

    def _locate_output_in_storage(self, task_id: TaskID) -> Optional[TaskID]:
        result = self.inspect_task(task_id)
        while isinstance(result.output_task_id, str):
            task_id = result.output_task_id
            result = self.inspect_task(result.output_task_id)
        if result.output_object_valid:
            return task_id
        return None

    def inspect_output(self, task_id: TaskID) -> Optional[TaskID]:
        """Get the actual checkpointed output for a task, represented by the ID of
        the task that actually keeps the checkpoint.

        Raises:
            ValueError: The workflow does not exist or the workflow state is not valid.

        Args:
            task_id: The ID of the task we are looking for its checkpoint.

        Returns:
            The ID of the task that actually keeps the checkpoint.
                'None' if the checkpoint does not exist.
        """
        status = self.load_workflow_status()
        if status == WorkflowStatus.NONE:
            raise ValueError(f"No such workflow '{self._workflow_id}'")
        if status == WorkflowStatus.CANCELED:
            raise ValueError(f"Workflow {self._workflow_id} is canceled")
        # For resumable workflow, the workflow result is not ready.
        # It has to be resumed first.
        if status == WorkflowStatus.RESUMABLE:
            raise ValueError(
                f"Workflow {self._workflow_id} is in resumable status, please resume it"
            )
        if task_id is None:
            task_id = self.get_entrypoint_task_id()
        return self._locate_output_in_storage(task_id)

    def inspect_task(self, task_id: TaskID) -> TaskInspectResult:
        """
        Get the status of a workflow task. The status indicates whether
        the workflow task can be recovered etc.

        Args:
            task_id: The ID of a workflow task

        Returns:
            The status of the task.
        """
        return self._inspect_task(task_id)

    def _inspect_task(self, task_id: TaskID) -> TaskInspectResult:
        items = self._scan(self._key_task_prefix(task_id), ignore_errors=True)
        keys = set(items)
        # does this task contains output checkpoint file?
        if STEP_OUTPUT in keys:
            return TaskInspectResult(output_object_valid=True)
        # do we know where the output comes from?
        if STEP_OUTPUTS_METADATA in keys:
            output_task_id = self._locate_output_task_id(task_id)
            return TaskInspectResult(output_task_id=output_task_id)

        # read inputs metadata
        try:
            metadata = self._get(self._key_task_input_metadata(task_id), True)
            return TaskInspectResult(
                args_valid=(STEP_ARGS in keys),
                func_body_valid=(STEP_FUNC_BODY in keys),
                workflow_refs=metadata["workflow_refs"],
                task_options=WorkflowTaskRuntimeOptions.from_dict(
                    metadata["task_options"]
                ),
                task_raised_exception=(STEP_EXCEPTION in keys),
            )
        except Exception:
            return TaskInspectResult(
                args_valid=(STEP_ARGS in keys),
                func_body_valid=(STEP_FUNC_BODY in keys),
                task_raised_exception=(STEP_EXCEPTION in keys),
            )

    def _save_object_ref(self, identifier: str, obj_ref: ray.ObjectRef):
        data = ray.get(obj_ref)
        self._put(self._key_obj_id(identifier), data)

    def load_actor_class_body(self) -> type:
        """Load the class body of the virtual actor.

        Raises:
            DataLoadError: if we fail to load the class body.
        """
        return self._get(self._key_class_body())

    def save_actor_class_body(self, cls: type) -> None:
        """Save the class body of the virtual actor.

        Args:
            cls: The class body used by the virtual actor.

        Raises:
            DataSaveError: if we fail to save the class body.
        """
        self._put(self._key_class_body(), cls)

    def save_task_prerun_metadata(self, task_id: TaskID, metadata: Dict[str, Any]):
        """Save pre-run metadata of the current task.

        Args:
            task_id: ID of the workflow task.
            metadata: pre-run metadata of the current task.

        Raises:
            DataSaveError: if we fail to save the pre-run metadata.
        """

        self._put(self._key_task_prerun_metadata(task_id), metadata, True)

    def save_task_postrun_metadata(self, task_id: TaskID, metadata: Dict[str, Any]):
        """Save post-run metadata of the current task.

        Args:
            task_id: ID of the workflow task.
            metadata: post-run metadata of the current task.

        Raises:
            DataSaveError: if we fail to save the post-run metadata.
        """

        self._put(self._key_task_postrun_metadata(task_id), metadata, True)

    def save_workflow_user_metadata(self, metadata: Dict[str, Any]):
        """Save user metadata of the current workflow.

        Args:
            metadata: user metadata of the current workflow.

        Raises:
            DataSaveError: if we fail to save the user metadata.
        """

        self._put(self._key_workflow_user_metadata(), metadata, True)

    def load_task_metadata(self, task_id: TaskID) -> Dict[str, Any]:
        """Load the metadata of the given task.

        Returns:
            The metadata of the given task.
        """

        def _load_task_metadata():
            if not self._scan(self._key_task_prefix(task_id), ignore_errors=True):
                if not self._scan("", ignore_errors=True):
                    raise ValueError(
                        "No such workflow_id '{}'".format(self._workflow_id)
                    )
                else:
                    raise ValueError(
                        "No such task_id '{}' in workflow '{}'".format(
                            task_id, self._workflow_id
                        )
                    )

            tasks = [
                self._get(self._key_task_input_metadata(task_id), True, True),
                self._get(self._key_task_prerun_metadata(task_id), True, True),
                self._get(self._key_task_postrun_metadata(task_id), True, True),
            ]

            (
                (input_metadata, _),
                (prerun_metadata, _),
                (postrun_metadata, _),
            ) = tasks

            input_metadata = input_metadata or {}
            prerun_metadata = prerun_metadata or {}
            postrun_metadata = postrun_metadata or {}

            metadata = input_metadata
            metadata["stats"] = {**prerun_metadata, **postrun_metadata}

            return metadata

        return _load_task_metadata()

    def load_workflow_metadata(self) -> Dict[str, Any]:
        """Load the metadata of the current workflow.

        Returns:
            The metadata of the current workflow.
        """

        def _load_workflow_metadata():
            if not self._scan("", ignore_errors=True):
                raise ValueError("No such workflow_id '{}'".format(self._workflow_id))

            tasks = [
                self._get(self._key_workflow_metadata(), True, True),
                self._get(self._key_workflow_user_metadata(), True, True),
                self._get(self._key_workflow_prerun_metadata(), True, True),
                self._get(self._key_workflow_postrun_metadata(), True, True),
            ]

            (
                (status_metadata, _),
                (user_metadata, _),
                (prerun_metadata, _),
                (postrun_metadata, _),
            ) = tasks

            status_metadata = status_metadata or {}
            user_metadata = user_metadata or {}
            prerun_metadata = prerun_metadata or {}
            postrun_metadata = postrun_metadata or {}

            metadata = status_metadata
            metadata["user_metadata"] = user_metadata
            metadata["stats"] = {**prerun_metadata, **postrun_metadata}

            return metadata

        return _load_workflow_metadata()

    def list_workflow(
        self, status_filter: Optional[Set[WorkflowStatus]] = None
    ) -> List[Tuple[str, WorkflowStatus]]:
        """List all workflows matching a given status filter.

        Args:
            status_filter: If given, only returns workflow with that status. This can
                be a single status or set of statuses.
        """
        return self._status_storage.list_workflow(status_filter)

    def delete_workflow(self) -> None:
        # TODO (Alex): There's a race condition here if someone tries to
        # start the workflow between these ops.
        self._status_storage.delete_workflow_status(self._workflow_id)
        found = self._storage.delete_dir("")
        # TODO (Alex): Different file systems seem to have different
        # behavior when deleting a prefix that doesn't exist, so we may
        # need to catch a broader class of exceptions.

        if not found:
            raise WorkflowNotFoundError(self._workflow_id)

    def update_workflow_status(self, status: WorkflowStatus):
        """Update the status of the workflow.
        This method is NOT thread-safe. It is handled by the workflow management actor.
        """
        self._status_storage.update_workflow_status(self._workflow_id, status)
        if status == WorkflowStatus.RUNNING:
            self._put(
                self._key_workflow_prerun_metadata(), {"start_time": time.time()}, True
            )
        elif status in (WorkflowStatus.SUCCESSFUL, WorkflowStatus.FAILED):
            self._put(
                self._key_workflow_postrun_metadata(), {"end_time": time.time()}, True
            )

    def load_workflow_status(self):
        """Load workflow status. If we find the previous status updating failed,
        fix it with redo-log transaction recovery."""
        return self._status_storage.load_workflow_status(self._workflow_id)

    def _put(self, key: str, data: Any, is_json: bool = False) -> str:
        """Serialize and put an object in the object store.

        Args:
            key: The key of the object.
            data: The data to be stored.
            is_json: If true, json encode the data, otherwise pickle it.
        """
        # TODO(suquark): Currently put to file is not atomic -- you can get a partial
        # file. This could fail workflow recovery.
        try:
            if not is_json:
                serialization.dump_to_storage(
                    key, data, self._workflow_id, storage=self
                )
            else:
                serialized_data = json.dumps(data).encode()
                self._storage.put(key, serialized_data)
        except Exception as e:
            raise DataSaveError from e

        return key

    def _get(self, key: str, is_json: bool = False, no_exception: bool = False) -> Any:
        err = None
        ret = None
        try:
            unmarshaled = self._storage.get(key)
            if unmarshaled is None:
                raise KeyNotFoundError
            if is_json:
                ret = json.loads(unmarshaled.decode())
            else:
                ret = cloudpickle.loads(unmarshaled)
        except KeyNotFoundError as e:
            err = e
        except Exception as e:
            err = DataLoadError()
            err.__cause__ = e

        if no_exception:
            return (ret, err)
        elif err is None:
            return ret
        else:
            raise err

    def _scan(self, prefix: str, ignore_errors: bool = False) -> List[str]:
        try:
            return [p.base_name for p in self._storage.list(prefix)]
        except Exception as e:
            if ignore_errors:
                return []
            raise e

    def _exists(self, key: str) -> bool:
        return self._storage.get_info(key) is not None

    # The following functions are helper functions to get the key
    # for a specific fields

    def _key_task_input_metadata(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_INPUTS_METADATA)

    def _key_task_user_metadata(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_USER_METADATA)

    def _key_task_prerun_metadata(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_PRERUN_METADATA)

    def _key_task_postrun_metadata(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_POSTRUN_METADATA)

    def _key_task_output(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_OUTPUT)

    def _key_task_exception(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_EXCEPTION)

    def _key_task_output_metadata(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_OUTPUTS_METADATA)

    def _key_task_function_body(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_FUNC_BODY)

    def _key_task_args(self, task_id):
        return os.path.join(STEPS_DIR, task_id, STEP_ARGS)

    def _key_obj_id(self, object_id):
        return os.path.join(OBJECTS_DIR, object_id)

    def _key_task_prefix(self, task_id):
        return os.path.join(STEPS_DIR, task_id, "")

    def _key_class_body(self):
        return os.path.join(CLASS_BODY)

    def _key_workflow_metadata(self):
        return os.path.join(WORKFLOW_META)

    def _key_workflow_user_metadata(self):
        return os.path.join(WORKFLOW_USER_METADATA)

    def _key_workflow_prerun_metadata(self):
        return os.path.join(WORKFLOW_PRERUN_METADATA)

    def _key_workflow_postrun_metadata(self):
        return os.path.join(WORKFLOW_POSTRUN_METADATA)

    def _key_num_tasks_with_name(self, task_name):
        return os.path.join(DUPLICATE_NAME_COUNTER, task_name)


def get_workflow_storage(workflow_id: Optional[str] = None) -> WorkflowStorage:
    """Get the storage for the workflow.

    Args:
        workflow_id: The ID of the storage.

    Returns:
        A workflow storage.
    """
    if workflow_id is None:
        workflow_id = workflow_context.get_workflow_task_context().workflow_id
    return WorkflowStorage(workflow_id)


def _load_object_ref(paths: List[str], wf_storage: WorkflowStorage) -> ObjectRef:
    @ray.remote(num_cpus=0)
    def load_ref(paths: List[str], wf_storage: WorkflowStorage):
        return wf_storage._get(paths)

    return load_ref.remote(paths, wf_storage)


@ray.remote(num_cpus=0)
def _put_obj_ref(ref: Tuple[ObjectRef]):
    """
    Return a ref to an object ref. (This can't be done with
    `ray.put(obj_ref)`).

    """
    return ref[0]
